/*
 * arena.cpp
 *
 * Arena and related classes
 *
 * Ulrik Petersen
 * Created: 3/2-2005
 * Last update: 8/26-2006
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2006  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

/**@file arena.cpp
 *@brief Implementation of Arena and related classes (EMdF layer)
 */

#include <arena.h>
#include <debug.h>

#include <iostream>

////////////////////////////////////////////////////////////////////
//
// ArenaConstIterator
//
////////////////////////////////////////////////////////////////////

/** Constructor.
 *
 * Constructs an empty ArenaConstIterator which is not associated with
 * any Arena.
 */
ArenaConstIterator::ArenaConstIterator()
{
	m_current = 0;
	m_index = 0;
	m_pMotherArena = 0;
	m_size = 0;
}


/** Constructor. Sets size and mother Arena.
 *
 * @param size The size in bytes of the elements in the Arena.
 *
 * @param pMotherArena The mother arena over which to iterate.
 */
ArenaConstIterator::ArenaConstIterator(unsigned int size, const Arena *pMotherArena)
{
	m_current = pMotherArena->m_head;
	m_index = 0;
	m_pMotherArena = pMotherArena;
	m_size = size;
}


/** Copy constructor.
 *
 * @param other The other ArenaConstIterator from which to initialize.
 */
ArenaConstIterator::ArenaConstIterator(const ArenaConstIterator& other)
{
	assign(other);
}


/** Destructor
 */
ArenaConstIterator::~ArenaConstIterator()
{
}


/** Check whether there are more elements in the Arena.
 *
 * @return True iff there is at least one more element of the specified size
 * in the Arena.
 */
bool ArenaConstIterator::hasNext() const
{
	if (m_current == 0) {
		return false;
	} else if (m_current->m_next != 0) {
		return true;
	} else if (m_current->m_index > m_index) {
		return true;
	} else {
		return false;
	}
}

/** Get pointer to current element and advance iterator.
 *
 * @return The pointer to the current element.
 */
void* ArenaConstIterator::next()
{
	ASSERT_THROW(hasNext(),
		     "hasNext() is false");
	void* pMem = (void*) (m_current->m_bytes + m_index);
	m_index += m_size;
	if (m_index >= Arena::Chunk::chunk_size
	    || m_index >= m_current->m_index) {
		m_current = m_current->m_next;
		m_index = 0;
	}
	return pMem;
}


/** Get pointer to current element but DO NOT advance iterator.
 *
 * @return The pointer to the current element.
 */
void* ArenaConstIterator::current()
{
	ASSERT_THROW(hasNext(), "hasNext() is false");
	void* pMem = m_current->m_bytes + m_index;
	return pMem;
}

/** Assignment operator.
 * 
 * @param other The other ArenaConstIterator from which to assign to
 * ourselves.
 * 
 * @return ourselves.
 */
ArenaConstIterator& ArenaConstIterator::operator=(const ArenaConstIterator& other)
{
	assign(other);
	return *this;
}


/** Assignment method.
 *
 * @param other The other ArenaConstIterator from which to assign to
 * ourselves.
 */
void ArenaConstIterator::assign(const ArenaConstIterator& other)
{
	m_current = other.m_current;
	m_index = other.m_index;
	m_pMotherArena = other.m_pMotherArena;
	m_size = other.m_size;
}



////////////////////////////////////////////////////////////
//
// Arena
//
////////////////////////////////////////////////////////////
/** Return an ArenaConstIterator over elements of the given size.
 *
 * @param size The size of the elements in the Arena.
 *
 * @return The ArenaConstIterator over this Arena with the specified
 * size element.
 */
ArenaConstIterator Arena::const_iterator(unsigned int size) const
{
	return ArenaConstIterator(size, this); 
}




