/*
 * mysqlconn.cpp
 *
 * MySQL Database connection 
 *
 * Ulrik Petersen
 * Created: 1/27-2001
 * Last update: 9/23-2006
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2006  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

/**@file mysqlconn.cpp
 *@brief Implementation of MySQL connection (EMdF layer)
 */


#ifdef _WIN32
#include <windows.h>
#endif

#include <debug.h>
#include <mysqlconn.h>
#include <string_func.h>
#include <iostream>
#include <cstring>
#include <cstdlib>
#include <logging.h>
#include <encryption.h>


MySQLEMdFConnection::MySQLEMdFConnection(const std::string& host, const std::string& user, const std::string& passwd, const std::string& db)
{
	// We aren't doing any transaction
	bTransactionInProgress = false;

	// We don't have any result
	pResultSet = NULL;
	m_Row = NULL;

	// We dont' have any connection
	pConn = NULL;

	// Get db
	if (db.length() == 0) {
		m_db = "emdf";
	} else {
		convertDBName(db, m_db);
	}

	// Set m_host, m_user, m_passwd
	m_host = host;
	m_user = user;
	m_passwd = passwd;

	// Encrypt password
	encryptString(m_passwd);

	// connect using the member variables just set
	connect();
}

MySQLEMdFConnection::~MySQLEMdFConnection()
{
	shutdownConnection();
}

void MySQLEMdFConnection::shutdownConnection()
{
	if (pConn != NULL) {
		// Commit transaction if in progress
		if (bTransactionInProgress)
			commitTransaction();
    
		// Do away with result set
		freeResult();
    
		// Close the connection and deallocate the MYSQL structure
		mysql_close(pConn);

		// Just to be safe...
		pConn = NULL;
	}
}


// Status of connection
// Returns true on connection good, false on connection bad
// Should be called right after constructor to see if we
// succeeded in setting up the connection.
bool MySQLEMdFConnection::connectionOk(void)
{
	if (pConn == NULL)
		return false;
	else {
		return mysql_ping(pConn) == 0;
	}
}


// Executing SELECT
bool MySQLEMdFConnection::execSelect(const std::string& query)
{
	if (pConn == 0) {
		return false;
	} else {
		// Do away with previous pResultSet
		freeResult();
    
		// Write to log
		LOG_WRITE_TIME("MySQLEMdFConnection::execSelect", std::string("Starting query.\n") + query);
    
		// Execute query
		if (mysql_query(pConn, query.c_str()) != 0) {
			// Write to log
			LOG_WRITE_TIME("MySQLEMdFConnection::execSelect", "Query finished with failure.");

			// Error: mysql_query did not return 0.
			return false;
		} else {
			// No error

			// Write to log
			LOG_WRITE_TIME("MySQLEMdFConnection::execSelect", "Query finished with success.");

			// Use result
			LOG_WRITE_TIME("MySQLEMdFConnection::execSelect", "Using result...");
			pResultSet = mysql_use_result(pConn);
			LOG_WRITE_TIME("MySQLEMdFConnection::execSelect", "Using result finished.");
			if (mysql_errno(pConn) != 0) {
				LOG_WRITE_TIME("MySQLEMdFConnection::execSelect", "Using result failed.");
				// An error occurred
				return false;
			} else {
				// No error occurred

				// Get first row
				bool bMoreRows;
				if (!getNextTuple(bMoreRows)) {
					// An error occurred
					return false;
				}

				// Get number of fields
				m_num_fields = mysql_field_count(pConn);

				// Return success
				return true;
			} 
		}
	}
}

// Executing all others but SELECT
bool MySQLEMdFConnection::execCommand(const std::string& query)
{
	if (pConn == 0) {
		return false;
	} else {
		// Do away with previous pResultSet
		freeResult();
    
		// Write to log
		LOG_WRITE_TIME("MySQLEMdFConnection::execCommand", std::string("Starting query.\n") + query);
    
		// Execute query
		if (mysql_query(pConn, query.c_str()) != 0) {
			// Write to log
			LOG_WRITE_TIME("MySQLEMdFConnection::execCommand", "Query finished with failure.");

			// Error: mysql_query did not return 0.
			return false;
		} else {
			// No error

			// "Store result" to make sure we've finished
			pResultSet = mysql_store_result(pConn);

			// This should be 0; it's a command...
			if (pResultSet != NULL) {
				// It wasn't 0, it is an error

				// Write to log
				LOG_WRITE_TIME("MySQLEMdFConnection::execCommand", "Query finished with success, BUT there were results.  This should not be, since this is a command.");

				// Append local error
				appendLocalError("MySQLEMdFConnection::execCommand: Query finished with success, BUT there were results.  This should not be, since this is a command.");

				// Free result
				mysql_free_result(pResultSet);
				pResultSet = NULL;

				// Return error
				return false;
			} else {
				// It was 0, there was no error

				// Write to log
				LOG_WRITE_TIME("MySQLEMdFConnection::execCommand", "Query finished with success.");
	
				// Get number of affected rows
				m_command_rows = mysql_affected_rows(pConn);
	
				// Return success
				return true;
			}
		} 
	}
}


// Get last error message
// Returns true on error message available 
// (even if it is "OK"), false on error message not available.
bool MySQLEMdFConnection::errorMessage(std::string& out)
{
	if (pConn == 0) {
		out = "not connected.";
		return true;
	} else {
		out = mysql_error(pConn);

		if (out.length() > 0)
			return true;
		else
			return false;
	}
}


// Transactions
bool MySQLEMdFConnection::beginTransaction(void)
{
	if (pConn == 0)
		return false;
	else {
		return false; // Transactions not supported
	}
}

bool MySQLEMdFConnection::commitTransaction(void)
{
	if (pConn == 0)
		return false;
	else {
		bTransactionInProgress = false;
		return false;
	}
}

bool MySQLEMdFConnection::abortTransaction(void)
{
	if (pConn == 0)
		return false;
	else {
		std::cerr << "Warning: MySQLEMdFConnection::abortTransaction: Transactions not supported." << std::endl << "... Something is wrong in the logic! ..." << std::endl;
		return false;
	}
}


// Field count
// Returns -1 on no fields.
int MySQLEMdFConnection::getNoOfFields(void)
{
	if (pConn == 0)
		return -1;
	else {
		if (pResultSet == NULL)
			return -1;
		else {
			return (int) mysql_num_fields(pResultSet);
		}
	}
}


// key is ignored
bool MySQLEMdFConnection::useDatabase(const std::string& db_name, const std::string& key)
{
	if (pConn == 0)
		return false;
	else {
		if (strcmp_nocase(db_name, "emdf") == 0) {
			shutdownConnection();
			m_db = "emdf";
			return connect();
		} else {
			int result = mysql_select_db(pConn, db_name.c_str());
			freeResult();
			return result == 0;
		}
	}
}

void MySQLEMdFConnection::freeResult(void)
{
	if (pResultSet != NULL) {
		// Fetch all remaining rows (if any)
		while (mysql_fetch_row(pResultSet) != NULL) ;

		// Free result
		mysql_free_result(pResultSet);

		// Clean up
		pResultSet = NULL;
		m_Row = NULL;
	}
}

std::string MySQLEMdFConnection::escapeStringForSQL(const std::string& str)
{
	// Malloc a piece of memory that is twice the input length + 1
	// as the MySQL manual demands.
	char *strTo = (char*) malloc(str.length()*2 + 1);

	// Call MySQL's API function.
	// This fills the buffer strTo with the scaped version of 
	// the third argument, and processes up to the number of 
	// bytes in the fourth argument.
	// This can even encode '\0'.
	// The return value is the number of bytes in the strTo buffer,
	// excluding the extra zero byte which the buffer is padded with.
	unsigned long newlength;
	newlength = mysql_real_escape_string(pConn, 
					     strTo, str.c_str(), str.length());

	// This std::string constructor is described on page 585 in
	// Stroustrup, TCPPPL, 3rd ed.
	std::string result = "'" + std::string(strTo, newlength) + "'";

	// Free the memory, since it has been copied to the result
	free(strTo);

	return result;

}

bool MySQLEMdFConnection::connect()
{
	// Make sure there isn't anything in progress
	shutdownConnection();

	// Make MYSQL structure
	pConn = mysql_init(NULL);

	// Read default configuration information
	mysql_options(pConn, MYSQL_READ_DEFAULT_GROUP, "client");

	// Decide DB name
	const char *pDBName = NULL;
	if (strcmp_nocase(m_db, "emdf") == 0) {
		// If the db is "emdf", don't set the default database when connecting
		pDBName = NULL;
	} else {
		pDBName = m_db.c_str();
	}
  
	// Get unencrypted form of password
	std::string unencrypted_password = m_passwd;
	unencryptString(unencrypted_password);

	// Connect
	MYSQL *res;
	if ((res = mysql_real_connect(pConn, m_host.c_str(), m_user.c_str(), unencrypted_password.c_str(), pDBName, 0, NULL, 0)) != pConn) {
		std::cerr << "MySQLEMdFDBConnection::connect():\nError: could not connect.  Return value from mysql_real_connect() was " << res << "." << std::endl;
		std::cerr << "mysql_error says: '" << mysql_error(pConn) << "'" << std::endl;
	}

	// Guard against an attack of looking at the memory or core-dumping
	zeroFillString(unencrypted_password);


#ifdef EMDF_DEBUG
	if (!connectionOk()) {
		std::cerr << "mysqlconn.cpp: Connection not OK connecting to database '"
			  << db << "'." << std::endl;
		std::string error_message;
		bool res_err;
		res_err = errorMessage(error_message);
		std::cout << "Error message = '" << error_message << "'\n";
		std::cout << "res_err = " << res_err << "\n";
	}
#endif // EMDF_DEBUG

	// Return connection status
	return connectionOk();
}
