/*
 * qdxml.cpp
 *
 * Quick'n Dirty XML SAX-like parser
 *
 * Ulrik Petersen
 * Created: 4/4-2007
 * Last update: 9/20-2008
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2007-2008  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

#include <cstring>
#include <qdxml.h>
#include <cstring>


/*
 * This hack is necessary because:
 *
 * - Visual C++ doesn't support std::map<_K,_V>::at().
 *
 * - The AttributeMap formal parameter is always const (and rightly so!)
 * 
 * - We can't use [] because it is not const.
 *
 * - Hence, we need another method.  This seems cleanest.
 *
 */
std::string getAttribute(const AttributeMap& attrs, const std::string& key) throw(QDException)
{
	AttributeMap::const_iterator ci = attrs.find(key);
	if (ci == attrs.end()) {
		throw QDException("Could not find attribute with name '" + key + "'");
	} else {
		return ci->second;
	}
}





const std::string QDParser::NOTAG = "<>";

std::string QDParser::state2string(eQDStates e)
{
	if (e == TEXT) {
		return "TEXT";
	} else if (e == ENTITY) {
		return "ENTITY";
	} else if (e == OPEN_TAG) {
		return "OPEN_TAG";
	} else if (e == CLOSE_TAG) {
		return "CLOSE_TAG";
	} else if (e == START_TAG) {
		return "START_TAG";
	} else if (e == ATTRIBUTE_LVALUE) {
		return "ATTRIBUTE_LVALUE";
	} else if (e == ATTRIBUTE_EQUAL) {
		return "ATTRIBUTE_EQUAL";
	} else if (e == ATTRIBUTE_RVALUE) {
		return "ATTRIBUTE_RVALUE";
	} else if (e == QUOTE) {
		return "QUOTE";
	} else if (e == IN_TAG) {
		return "IN_TAG";
	} else if (e == SINGLE_TAG) {
		return "SINGLE_TAG";
	} else if (e == COMMENT) {
		return "COMMENT";
	} else if (e == DONE) {
		return "DONE";
	} else if (e == DOCTYPE) {
		return "DOCTYPE";
	} else if (e == IN_XMLDECLARATION) {
		return "IN_XMLDECLARATION";
	} else if (e == BEFORE_XMLDECLARATION) {
		return "BEFORE_XMLDECLARATION";
	} else if (e == OPEN_XMLDECLARATION) {
		return "OPEN_XMLDECLARATION";
	} else if (e == CDATA) {
		return "CDATA";
	} else {
		std::ostringstream str;
		str << "Unknown state: " << (int) e;
		throw QDException(str.str());
	}
}
	


void QDParser::parse(QDDocHandler *pDH, std::istream *pStream) throw(QDException) 
{
	bot = tok = ptr = cur = pos = lim = top = eof = 0;
	m_pInStream = pStream;
	m_pDH = pDH;
	m_tag_depth = 0;
	eQDStates state = BEFORE_XMLDECLARATION;
	m_encoding = kCSUTF8; // This is the default for XML
	m_tag_name = NOTAG;
	m_line = 1;
	m_column = 0;
	m_pDH->startDocument();
	m_end_of_line = false;

	char c = getNextChar();

	while (hasMoreInput()) {
		//std::cerr << "UP200: state = " << state2string(state) << ", c = '" << c << "'" << std::endl;

		if (state == TEXT) {
			// We are collecting text between tags.
			if (c == '&') {
				pushState(TEXT);
				m_entity = "";
				state = ENTITY;
			} else if (c == '<') {
				if (m_text.length() > 0) {
					m_pDH->text(m_text);
					m_text = "";
				}
				pushState(TEXT);
				state = START_TAG;
				m_tag_name = "";
			} else {
				m_text += c;
			}
		} else if (state == BEFORE_XMLDECLARATION) {
			if (c == '<') {
				c = getNextChar();
				if (c == '?') {
					c = getNextChar();
					if (!c == 'x') {
						except("Document must start with '<?x'. It does not.");
					}
					c = getNextChar();
					if (!c == 'm') {
						except("Document must start with '<?xm'. It does not.");
					}
					c = getNextChar();
					if (!c == 'l') {
						except("Document must start with '<?xml'. It does not.");
					}
					c = getNextChar();
					if (!myisspace(c)) {
						except("Document must start with '<?xml' + <whitespace>. It does not.");
					}
					state = IN_XMLDECLARATION;
				} else {
					except("Document must start with <?; it does not.");
				}
			}
		} else if (state == IN_XMLDECLARATION) {
			// We are inside <?xml .... ?>
			if (c == '?') {
				c = getNextChar();
				if (c != '>') {
					except("XML declaration must end with '?>'");
				} else {
					if (m_attributes.find("encoding") != m_attributes.end()) {
						std::string strencoding = m_attributes["encoding"];
						eCharsets tmpcharset;
						if (string2charset(strencoding, tmpcharset)) {
							m_encoding = tmpcharset;
						} else {
							// Use UTF-8 if charset unknown.
							// FIXME: Should we do an exception?
							m_encoding = kCSUTF8;
						}
					} else {
						m_encoding = kCSUTF8;
					}

					state = TEXT;
				}
			} else if (myisspace(c)) {
				; // Eat up
			} else {
				pushState(IN_XMLDECLARATION);
				state = ATTRIBUTE_LVALUE;
				m_lvalue = "";
				m_rvalue = "";
				m_lvalue += c;
			}
		} else if (state == ATTRIBUTE_LVALUE) {
			if (c == '=') {
				state = ATTRIBUTE_RVALUE;
			} else if (myisspace(c)) {
				state = ATTRIBUTE_EQUAL;
			} else {
				m_lvalue += c;
			}
		} else if (state == ATTRIBUTE_EQUAL) {
			if (c == '=') {
				state = ATTRIBUTE_RVALUE;
			} else if (myisspace(c)) {
				; // Eat up
			} else {
				except("Error while processing attribute: After lvalue, there should only be whitespace or '='.");
			}
		} else if (state == ATTRIBUTE_RVALUE) {
			if (c == '\'' || c == '\"') {
				m_quote_char = c;
				state = QUOTE;
			} else if (myisspace(c)) {
				; // Eat up
			} else {
				except("Error while processing attribute: After =, there should only be whitespace or '\\'' or '\\\"'.");
			}
		} else if (state == QUOTE) {
			if (c == '\t' || c == '\n' || c == '\r') {
				// See XML Spec section 3.3.3
				// on normalization.
				m_rvalue += ' ';
			} else if (c == '&') {
				pushState(QUOTE);
				state = ENTITY;
				m_entity = "";
			} else if (c == m_quote_char) {
				m_attributes.insert(AttributeMap::value_type(m_lvalue, m_rvalue));
				state = popState();
			} else {
				m_rvalue += c;
			}
		} else if (state == ENTITY) {
			if (c == ';') {
				state = popState();
				handle_entity(state);
			} else {
				m_entity += c;
			}
		} else if (state == IN_TAG) {
			if (c == '/') {
				state = SINGLE_TAG;
			} else if (c == '>') {
				++m_tag_depth;
				m_pDH->startElement(m_tag_name, m_attributes);
				eraseAttributes();
				m_tag_name = "";
				state = popState();
			} else if (myisspace(c)) {
				; // Eat up
			} else {
				pushState(IN_TAG);
				state = ATTRIBUTE_LVALUE;
				m_lvalue = "";
				m_rvalue = "";
				m_lvalue += c;
			}
		} else if (state == START_TAG) {
			// The last char was '<'. Now
			// find out whether we are doing <starttag>,
			// </endtag>, <!-- comment -->,
			// <!DOCTYPE, or <![CDATA
			eQDStates topState = popState();
			if (c == '/') {
				pushState(topState);
				state = CLOSE_TAG;
			} else if (c == '?') {
				pushState(topState);
				state = OPEN_XMLDECLARATION;
				eraseAttributes();
			} else {
				pushState(topState);
				state = OPEN_TAG;
				m_tag_name = c;
				eraseAttributes();
			}
		} else if (state == OPEN_TAG) {
			if (c == '/') {
				state = SINGLE_TAG;
			} else if (c == '>') {
				++m_tag_depth;
				m_pDH->startElement(m_tag_name, m_attributes);
				eraseAttributes();
				m_tag_name = "";
				state = popState();
			} else if (myisspace(c)) {
				state = IN_TAG;
			} else {
				m_tag_name += c;
				if (c == '-' && m_tag_name == "!--") {
					pushState(TEXT);
					state = COMMENT;
				} else if (c == '[' && m_tag_name == "![CDATA[") {
					m_text = "";
					pushState(TEXT);
					state = CDATA;
				} else if (c == 'E' && m_tag_name == "!DOCTYPE") {
					pushState(TEXT);
					state = DOCTYPE;
					m_text = "";
					m_tag_name = "";
				} 
			}
		} else if (state == COMMENT) {
			if (c == '>') {
				m_text += c;
				if (string_ends_with(m_text, "-->")) {
					m_pDH->comment(m_text);

					m_text = "";
					state = popState();
				} else {
					; // Nothing to do
				}
			} else {
				m_text += c;
			}
		} else if (state == CDATA) {
			if (c == '>') {
				if (string_ends_with(m_text, "]]")) {
					m_pDH->text(m_text.substr(0, m_text.length() - 2));
					state = popState();
					m_text = "";
				} else {
					m_text += c;
				}
			} else {
				m_text += c;
			}
		} else if (state == DOCTYPE) {
			m_tag_name += c;
			if (c == '<') {
				m_tag_name = "";
			} else if (c == '-' && m_tag_name == "!--") {
				pushState(DOCTYPE);
				state = COMMENT;
			} else if (c == ']') {
				m_tag_name = c;
			} else if (c == '>') {
				if (m_tag_name == "]>") {
					// Is TEXT
					state = popState();
				}
			} else {
				; // Eat up
			}
		} else if (state == CLOSE_TAG) {
			// We have seen '</'
			if (c == '>') {
				m_pDH->endElement(m_tag_name);
				state = popState();
				--m_tag_depth;
				if (m_tag_depth == 0) {
					state = DONE;
				}
				m_tag_name = "";
			} else {
				m_tag_name += c;
			}
			
		} else if (state == SINGLE_TAG) {
			// We are have seen 
			// something like
			// <barefoot attr1="value" ... /
			// Now we need the final >
			if (c != '>') {
				except("While processing single tag: Expected > for tag: <" + m_tag_name + "/>");
			} else {
				m_pDH->startElement(m_tag_name, m_attributes);
				m_pDH->endElement(m_tag_name);
				if (m_tag_depth == 0) {
					state = DONE;
				} else {
					state = popState();
					m_tag_name = "";
					eraseAttributes();
				}
			}
		} else if (state == DONE) {
			eraseAttributes();
			m_pDH->endDocument();
			cleanUp();
			return;
		} else {
			except(std::string("Error: Unknown state: ") + state2string(state));
		}

		// Read next char
		c = getNextChar();	
	}
	
	if (state == DONE) {
		m_pDH->endDocument();
	} else {
		except("missing end tag");
	}
	eraseAttributes();
	cleanUp();
}
	
void QDParser::except(const std::string& s) throw(QDException)
{
	std::ostringstream ostr;
	ostr << s << " near line " << m_line << ", column " << m_column << '!' << std::endl;
	throw QDException(ostr.str());
}

void QDParser::handle_entity(eQDStates state) throw(QDException)
{
	if (m_entity.length() > 0 && m_entity[0] == '#') {
		long l = 0;
		if (m_entity.length() > 1
		    && (m_entity[1] == 'x' 
			|| m_entity[1] == 'X')) {
			std::string hexstring = m_entity.substr(2, std::string::npos);
			if (is_hex(hexstring)) {
				l = hex2long(hexstring);
			} else {
				except("Invalid hex-string in entity &" + m_entity + ";.");
			}
		} else {
			std::string decimalstring = m_entity.substr(1, std::string::npos);
			l = string2long(decimalstring);
		}

		if (m_encoding == kCSUTF8) {
			std::string utf8str;
			long2utf8(l, utf8str);
			std::string::size_type i;
			std::string::size_type utf8length = utf8str.length();
			for (i = 0; i < utf8length; ++i) {
				if (state == QUOTE) {
					m_rvalue += utf8str[i];
				} else {
					m_text += utf8str[i];
				}
			}
		} else {
			if (l <= 255) {
				if (state == QUOTE) {
					m_rvalue += (char) l;
				} else {
					m_text += (char) l;
				}
			} else {
				except("Entity &" + m_entity + "; out of range; document is non-UTF8, and so we can handle only single-byte data.");
			}
		}
	} else {
		char entity_char = '\0';
		if (m_entity == "lt") {
			entity_char = '<';
		} else if (m_entity == "gt") {
			entity_char = '>';
		} else if (m_entity == "amp") {
			entity_char = '&';
		} else if (m_entity == "quot") {
			entity_char = '"';
		} else if (m_entity == "apos") {
			entity_char = '\'';
		} else {
			except("Unknown entity: &"+m_entity+";");
		}
		
		if (state == QUOTE) {
			m_rvalue += entity_char;
		} else {
			m_text += entity_char;
		}
	}
}


char QDParser::getNextChar(void) throw(QDException)
{
	char c;
	if ((lim - cur) < 1) fillBuffer();
	c = *cur;
	++cur;
	if (c == '\r') {
		m_end_of_line = true;
		m_column = 0;
		++m_line;
		c = '\n';
	} else if (m_end_of_line && c == '\n') {
		// We have already read \r, and this is \n that follows.
		// \r returned \n, since we need to translate
		// all of \r\n, \r, and \n to \n 
		// (see Section 2.11 of the XML spec)
		// So now we return the next character
		m_pInStream->get(c);
		m_end_of_line = false;
	} else if (c == '\n') {
		m_column = 0;
		++m_line;
	} else {
		m_end_of_line = false; // We might be doing a file with \r as the end-of-line marker.
		++m_column;
	}
	return c;
}

#define BSIZE   (1024*1024)


void QDParser::fillBuffer()
{
        if(!eof)
        {
                unsigned int cnt = tok - bot;
		/*
		std::cerr << "UP205: "
			  << " cnt = " << cnt 
			  << " tok = " << (void*) tok
			  << " bot = " << (void*) bot 
			  << " cur = " << (void*) cur
			  << " eof = " << (void*) eof 
			  << " ptr = " << (void*) ptr
			  << " lim = " << (void*) lim
			  << std::endl;
		*/
                if(cnt)
                {
                        memcpy(bot, tok, lim - tok);
                        tok = bot;
                        ptr -= cnt;
                        cur -= cnt;
                        pos -= cnt;
                        lim -= cnt;
                }
                if((top - lim) < BSIZE)
                {
                        char *buf = new char[(lim - bot) + BSIZE];
                        memcpy(buf, tok, lim - tok);
                        tok = buf;
                        ptr = &buf[ptr - bot];
                        cur = &buf[cur - bot];
                        pos = &buf[pos - bot];
                        lim = &buf[lim - bot];
                        top = &lim[BSIZE];
                        delete [] bot;
                        bot = buf;
                }
                m_pInStream->read(lim, BSIZE);
                if ((cnt = m_pInStream->gcount()) != BSIZE )
                {
                        eof = &lim[cnt]; *eof++ = '\0';
                }
                lim += cnt;
        }
}

QDParser::QDParser()
{
	bot = tok = ptr = cur = pos = lim = top = eof = 0;
}

QDParser::~QDParser()
{
	cleanUp();
}
