/*
 * string_func.cpp
 *
 * String helper functions
 *
 * Ulrik Petersen
 * Created: 3/1-2001
 * Last update: 5/25-2008
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2008  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

/**@file string_func.cpp
 *@brief Implementation of string functions (EMdF layer)
 */

#include <cstdlib>
#include <string_func.h>
#include <string_list.h>
#include <debug.h>
#include <cctype>
#include <cstring>
#include <cstdlib>
#include <sstream>
#include <emdros_reserved_words.h>



/** Upper-case a string.
 *
 * @param in The input string (not changed).
 *
 * @param result The output string.
 */
void str_toupper(const std::string& in, std::string& result)
{
	result = "";
	std::string::const_iterator ci(in.begin());
	std::string::const_iterator cend(in.end());
	while (ci != cend) {
		result += (unsigned char) (toupper(*ci));
		++ci;
	}
}

/** Lower-case a string.
 *
 * @param in The input string (not changed).
 *
 * @param result The output string.
 */
void str_tolower(const std::string& in, std::string& result)
{
	result = "";
	std::string::const_iterator ci(in.begin());
	std::string::const_iterator cend(in.end());
	while (ci != cend) {
		result += (unsigned char) (tolower(*ci));
		++ci;
	}
}

/** Compare two strings case-INsensitively.
 *
 * See Bjarne Stroustrup, 
 * The C++ Programming Language,
 * 3rd Edition, 1997
 * p. 591
 *
 * The standard C "tolower" function is used to evaluate
 * string-sensitivity.
 *
 * @param str1 The first string.
 *
 * @param str2 The second string.
 *
 * @return 0 if they are equal, -1 if str1 < str2, 1 if str1 > str2,
 * all less case-sensitivity.
 */

int strcmp_nocase(const std::string& str1, const std::string& str2)
{
	std::string::const_iterator p = str1.begin();
	std::string::const_iterator str1end = str1.end();
	std::string::const_iterator q = str2.begin();
	std::string::const_iterator str2end = str2.end();
	while (p != str1end && q != str2end) {
		unsigned char tolop = tolower(*p);
		unsigned char toloq = tolower(*q);
		if (tolop != toloq)
			return (tolop < toloq) ? -1 : 1;
		++p;
		++q;
	}
	std::string::size_type str1size = str1.size();
	std::string::size_type str2size = str2.size();
	return (str2size == str1size) ? 0 : (str1size < str2size) ? -1 : 1;
}

/** Compare two strings with case-sensitivity.
 *
 * See Bjarne Stroustrup, 
 * The C++ Programming Language,
 * 3rd Edition, 1997
 * p. 591
 *
 * @param str1 The first string.
 *
 * @param str2 The second string.
 *
 * @return 0 if they are equal, -1 if str1 < str2, 1 if str1 > str2.
 */
int strcmp_case(const std::string& str1, const std::string& str2)
{
	std::string::const_iterator p = str1.begin();
	std::string::const_iterator str1end = str1.end();
	std::string::const_iterator q = str2.begin();
	std::string::const_iterator str2end = str2.end();
	while (p != str1end && q != str2end) {
		unsigned char cp = *p;
		unsigned char cq = *q;
		if (cp != cq)
			return (cp < cq) ? -1 : 1;
		++p;
		++q;
	}
	std::string::size_type str1size = str1.size();
	std::string::size_type str2size = str2.size();
	return (str2size == str1size) ? 0 : (str1size < str2size) ? -1 : 1;
}

/** Convert a bool to a string (either "0" or "1").
 *
 * @param b The bool to convert.
 *
 * @return "1" if b is true, "0" otherwise.
 */
std::string bool2string(bool b)
{
	if (b)
		return "1";
	else
		return "0";
}

/** Convert an eBackendKind enumeration constant to a string.
 *
 * @param backend_kind The backend kind to convert.
 *
 * @return Either "PostgreSQL", "MySQL", "SQLite", or "BackendNone"
 */
std::string backend_kind2string(eBackendKind backend_kind)
{
	switch (backend_kind) {
	case kPostgreSQL:
		return "PostgreSQL";
	case kMySQL:
		return "MySQL";
	case kSQLite2:
		return "SQLite";
	case kSQLite3:
		return "SQLite3";
	case kBackendNone:
		return "BackendNone";
	default:
		ASSERT_THROW(false, "Unknown backend.");

		// Just to fool the compiler into not giving a warning.
		return "Unknown backend."; 
	}

}

/** Convert a string to a bool (assuming it is either "0" or "1").
 *
 * @param str The input string.
 *
 * @return True if and only if str is "1".
 */
bool string2bool(const std::string& str)
{
	return str == "1";
}

/** Convert a boolean to one of the strings "true" or "false".
 *
 * @param b The bool to convert.
 *
 * @return If b is true, then "true", otherwise, "false".
 */
std::string bool_alpha2string(bool b)
{
	if (b)
		return "true";
	else
		return "false";
}

/** Convert a string to a bool.
 *
 * @param str The input string.
 *
 * @return True if and only if the input string is "true".
 */
bool string2bool_alpha(const std::string& str)
{
	return str == "true";
}

/** Convert an int to a string, base 10.
 *
 * @param i The int to convert.
 *
 * @return The input int, converted to a string.
 */
std::string int2string(int i)
{
	char szResult[30];
	sprintf(szResult, "%d", i);
	return szResult;
}

/** Convert a double to a string, base 10.
 *
 * @param d The double to convert.
 *
 * @return The input double, converted to a string, base 10.
 */
std::string double2string(double d)
{
	char szResult[60];
	sprintf(szResult, "%lf", d);
	return szResult;
}


/** Convert a monad_m to a string, base 10.
 *
 * @param m The monad_m to convert.
 *
 * @return The input monad, converted to a string, base 10.
 */
std::string monad_m2string(monad_m m)
{
	char szResult[30];
	sprintf(szResult, "%ld", (long) m);
	return szResult;
}

/** Convert a string to a monad_m.
 *
 * @param str The input string (having the monad in base-10).
 *
 * @return The monad_m represented by the string.
 */
monad_m string2monad_m(const std::string& str)
{
	return strtol(str.c_str(), (char **)NULL, 10);
}

/** Convert an id_d to a string, that has "NIL" for the numerical
 *  value of nil.
 *
 * @see id_d2number_string(id_d_t)
 *
 * @param i The id_d to convert.
 *
 * @return The input id_d as a string which has "NIL" for the
 * numerical value of nil.
 */
std::string id_d2string(id_d_t i)
{
	char szResult[30];
	if (i == NIL) {
		strcpy(szResult, NIL_AS_VISIBLE_STRING);
	} else {
		sprintf(szResult, "%ld", (long) i);
	}
	return szResult;
}

/** Convert an id_d to a string that does not contain "NIL" for the
 *  nil id_d, but instead its numerical value.
 *
 * @see id_d2string(id_d_t).
 *
 * @param i The id_d to convert.
 *
 * @return The input id_d as a numerical string.
 */
std::string id_d2number_string(id_d_t i)
{
	char szResult[30];
	sprintf(szResult, "%ld", (long) i);
	return szResult;
}

/** Convert a long to a string.
 *
 * @param l The long to convert.
 *
 * @return The input long, converted to a string (base 10).
 */
std::string long2string(long l)
{
	char szResult[30];
	sprintf(szResult, "%ld", l);
	return szResult;
}


/** Convert a string to a long.
 *
 * @param str The input string, base 10.
 *
 * @return The input string, converted to a long.
 */
long string2long(const std::string& str)
{
	return strtol(str.c_str(), (char **)NULL, 10);
}

/** Convert a string to an ID_D.
 *
 * "NIL" is converted to the numeric value of NIL.
 *
 * @param str The input string.
 *
 * @return The input string, converted to an id_d.
 */
id_d_t string2id_d(const std::string& str)
{
	id_d_t id_d;
	if (str == NIL_AS_VISIBLE_STRING) {
		id_d = NIL;
	} else {
		id_d = strtol(str.c_str(), (char **)NULL, 10);
	}
	return id_d;
}

/** Encode a string so it can be included in "double quotes" in an MQL
 *  script.
 *
 * @param str The input string (this is not changed).
 *
 * @return The input string, encoded so as to be able to be included
 * in an MQL script in "double quotes".
 */
std::string encodeSTRINGstring(const std::string& str)
{
	std::string result;
	for (std::string::size_type i = 0; i < str.length(); ++i) {
		unsigned char c = (unsigned char) str[i];
		if (c < 0x20 
		    || c >= 0x80) {
			result += "\\x" + char2hex(c);
		} else {
			switch (c) {
			case '\\':
				result += "\\\\";
				break;
			case '\"':
				result += "\\\"";
				break;
			default:
				result += c;
				break;
			}
		}
	}
	return result;
}

/** Finds out whether a string is anything other than whitespace.
 *
 * @param str The string to search (is not changed).
 *
 * @return True if and only if str consists solely of
 * whitespace-characters. (" \t\n\r\v").
 */
bool is_other_than_whitespace(const std::string& str)
{
	return str.find_first_not_of(std::string(" \t\n\r\v"), 0) != std::string::npos;
}

/** Finds out whether a string has a double quote (").
 *
 * @param str The string to search (is not changed).
 *
 * @return True if and only if str has a double quote.
 */
bool has_double_quote(const std::string& str)
{
	return str.find_first_of('"') != std::string::npos;
}

/** Checks whether string is an identifier
 *
 * An identifer starts with an underscore, a lowercase-letter (a-z) or
 * an upper-case letter (A-Z).  If it is longer than one character,
 * the rest consists of either the characters just mentioned or digits
 * (0-9).
 *
 * @param str the string to be checked for identifierhood.
 * @return true if str is an identifier, false otherwise.
 */
bool is_identifier(const std::string& str)
{
	std::string non_digit = "_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
	std::string digit = "0123456789";
	std::string allowed_chars = "_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
	if (str == "") {
		return false;
	} else {
		if (non_digit.find(str[0]) == std::string::npos) {
			return false;
		} else {
			if (str.find_first_not_of(allowed_chars) != std::string::npos) {
				return false;
			} else {
				return true;
			}
		}
	}
}

/** Check whether a string is a hexadecimal string.
 *
 * This just checks whether the string only contains [0-9A-Fa-f].
 *
 * @param str The string to check for hexadecimal-hood.
 *
 * @return True iff str is a hexadecimal string.
 */
bool is_hex(const std::string& str)
{
	const std::string hex_digits = "0123456789abcdefABCDEF";
	return str != "" && str.find_first_not_of(hex_digits) == std::string::npos;
}

/** Returns the long value of an arbitrary hex string.
 *
 * The input is assumed to have been sanity-checked before hand, e.g.,
 *  with is_hex().
 *
 * @param str The string to convert
 *
 * @return The long representing the input hex string.
 */
long hex2long(const std::string& str)
{
	long result = 0;
	std::string::size_type strlength = str.length();
	bool bHasMetFirstNonZero = false;
	for (std::string::size_type i = 0;
	     i < strlength;
	     ++i) {
		char c = str[i];
		// Skip leading zeros
		if (c != '0') {
			bHasMetFirstNonZero = true;
		}
		if (bHasMetFirstNonZero) {
			long tmp = 0;
			if (c >= '0'
			    && c <= '9') {
				tmp = c - '0';
			} else if (c >= 'a'
				   && c <= 'f') {
				tmp = c - ('a' - 10);
			} else if (c >= 'A'
				   && c <= 'F') {
				tmp = c - ('A' - 10);
			}
			result <<= 4;
			result |= tmp;
		}
	}
	return result;
}


/** Replace occurrences of one character with another character in a
 *  string.
 *
 * @param s The input (and output) string.
 *
 * @param in The character to change from.
 *
 * @param out The character to change to.
 */
void replace_chars(std::string &s, const char in, const char out)
{
	std::string::size_type i = 0;
	while ((i = s.find(in, i)) != std::string::npos)
		s[i] = out;
}

/** Replace underscores with spaces.
 *
 * @param s The input (and output) string.
 */
void replace_underscores(std::string& s)
{
	replace_chars(s, '_', ' ');
}


/** Replace spaces with underscores.
 *
 * @param s The input (and output) string.
 */
void replace_spaces(std::string& s)
{
	replace_chars(s, ' ', '_');
}

/** Reverse a string.
 *
 * @param in The input string (is not changed).
 *
 * @return The input string, reversed.
 */
std::string reverse_string(const std::string& in)
{
	std::string::size_type i;
	std::string result;
	if (in.length() > 0) {
		i = in.length();
		do {
			i--;
			result.append(1, in[i]);
		} while (i != 0);
	}

	return result;
}

/** Finds out whether a string is an integer or not.
 *
 * @param in The input string.
 *
 * @return True if this is an integer, False if it is not.
 */
bool string_is_number(const std::string& in)
{
	return (in.size() > 0) 
		&& (in.find_first_not_of("0123456789", 0) == std::string::npos);
}


/** Convert a string to URL-encoded (%-hex-encoded) strings.
 *
 * Only characters above 0x7F (and the %-character) are converted to
 * %xx-hex-format.
 * 
 * @param instr The input string (is not changed).
 *
 * @return The input string, converted to hex where necessary.
 */
std::string convert_to_hex(std::string instr)
{
	std::string outstr = "";
	for (unsigned int i = 0; i < instr.length(); i++) {
		if (isspace((unsigned char)instr[i]) 
		    || instr[i] == '%' 
		    || ((unsigned char) instr[i]) >= 0x80) {
			outstr.append(1, '%');

			// Get hexadecimal representation
			std::string hex;
			std::ostringstream ostream(hex);
			ostream.setf(std::ios_base::hex, std::ios_base::basefield);
			ostream.precision(2);
			ostream << (int) ((unsigned char) instr[i]);
			outstr.append(ostream.str());
		} else {
			outstr.append(1, instr[i]);
		}
	}
	return outstr;
}


/** Remove all instances of a given character from a string, and return
 * the result.
 *
 * @param instr The input string (is not changed).
 *
 * @param char_to_remove The character to remove.
 *
 * @return The input string, less any instances of the given character.*
 */
std::string remove_char(std::string instr, char char_to_remove)
{
	std::string outstr = "";
	for (unsigned int i = 0; i < instr.length(); i++) {
		if (instr[i] != char_to_remove) {
			outstr.append(1, instr[i]);
		}
	}
	return outstr;
}

/** Remove all instances of characters given string in a from a
 * string, and return the result.
 *
 * @param instr The input string (is not changed).
 *
 * @param chars_to_remove The characters to remove.
 *
 * @return The input string, less any instances of the given characters.
 */
std::string remove_chars_in_string(const std::string& instr, const std::string& chars_to_remove)
{
	std::string result;
	for (std::string::size_type i = 0;
	     i < instr.length();
	     ++i) {
		char c = instr[i];
		if (chars_to_remove.find(c, 0) != std::string::npos) {
			// skipt his char
		} else {
			// add this char
			result.append(1,c);
		}
	}
	return result;
}

/** Remove punctuation from a string.
 *
 * Punctuation is defined as being one of the characters
 * "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~".
 *
 * @param instr The input string (is not changed).

 * @return The input string, less any instances of the given
 * punctuation characters.
 */
std::string remove_punct(const std::string instr)
{
	return remove_chars_in_string(instr, "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~");
}


/** Convert a string to an eCharsets enumeration.
 *
 * Find out which eCharsets enumeration constant most closely matches
 * a given string.
 *
 *
 * @param input The string-representation (probably from the user)
 * which is to be converted to eCharsets.
 *
 * @param charset The output charset.  Is only defined if the function
 * returns true.
 *
 * @return True iff we were able to recognize and thus convert the
 * charset.
 *
 */
bool string2charset(const std::string& input, eCharsets& charset)
{
	if (strcmp_nocase(input, "ascii") == 0) {
		charset = kCSASCII;       /**< 7-bit ASCII.  */
		return true;
	} else if (strcmp_nocase(input, "utf8") == 0
		   || strcmp_nocase(input, "utf-8") == 0) {
 		charset = kCSUTF8;        /**< UTF-8. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-1") == 0
		   || strcmp_nocase(input, "iso_8859_1") == 0
		   || strcmp_nocase(input, "iso-8859_1") == 0
		   || strcmp_nocase(input, "iso-8859-1") == 0) {
		charset = kCSISO_8859_1;  /**< ISO-8859-1. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-2") == 0
		   || strcmp_nocase(input, "iso_8859_2") == 0
		   || strcmp_nocase(input, "iso-8859_2") == 0
		   || strcmp_nocase(input, "iso-8859-2") == 0) {
		charset = kCSISO_8859_2;  /**< ISO-8859-2. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-3") == 0
		   || strcmp_nocase(input, "iso_8859_3") == 0
		   || strcmp_nocase(input, "iso-8859_3") == 0
		   || strcmp_nocase(input, "iso-8859-3") == 0) {
		charset = kCSISO_8859_3;  /**< ISO-8859-3. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-4") == 0
		   || strcmp_nocase(input, "iso_8859_4") == 0
		   || strcmp_nocase(input, "iso-8859_4") == 0
		   || strcmp_nocase(input, "iso-8859-4") == 0) {
		charset = kCSISO_8859_4;  /**< ISO-8859-4. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-5") == 0
		   || strcmp_nocase(input, "iso_8859_5") == 0
		   || strcmp_nocase(input, "iso-8859_5") == 0
		   || strcmp_nocase(input, "iso-8859-5") == 0) {
		charset = kCSISO_8859_5;  /**< ISO-8859-5. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-6") == 0
		   || strcmp_nocase(input, "iso_8859_6") == 0
		   || strcmp_nocase(input, "iso-8859_6") == 0
		   || strcmp_nocase(input, "iso-8859-6") == 0) {
		charset = kCSISO_8859_6;  /**< ISO-8859-6. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-7") == 0
		   || strcmp_nocase(input, "iso_8859_7") == 0
		   || strcmp_nocase(input, "iso-8859_7") == 0
		   || strcmp_nocase(input, "iso-8859-7") == 0) {
		charset = kCSISO_8859_7;  /**< ISO-8859-7. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-8") == 0
		   || strcmp_nocase(input, "iso_8859_8") == 0
		   || strcmp_nocase(input, "iso-8859_8") == 0
		   || strcmp_nocase(input, "iso-8859-8") == 0) {
		charset = kCSISO_8859_8;  /**< ISO-8859-8. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-9") == 0
		   || strcmp_nocase(input, "iso_8859_9") == 0
		   || strcmp_nocase(input, "iso-8859_9") == 0
		   || strcmp_nocase(input, "iso-8859-9") == 0) {
		charset = kCSISO_8859_9;  /**< ISO-8859-9. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-10") == 0
		   || strcmp_nocase(input, "iso_8859_10") == 0
		   || strcmp_nocase(input, "iso-8859_10") == 0
		   || strcmp_nocase(input, "iso-8859-10") == 0) {
		charset = kCSISO_8859_10;  /**< ISO-8859-10. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-13") == 0
		   || strcmp_nocase(input, "iso_8859_13") == 0
		   || strcmp_nocase(input, "iso-8859_13") == 0
		   || strcmp_nocase(input, "iso-8859-13") == 0) {
		charset = kCSISO_8859_13;  /**< ISO-8859-13. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-14") == 0
		   || strcmp_nocase(input, "iso_8859_14") == 0
		   || strcmp_nocase(input, "iso-8859_14") == 0
		   || strcmp_nocase(input, "iso-8859-14") == 0) {
		charset = kCSISO_8859_14;  /**< ISO-8859-14. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-15") == 0
		   || strcmp_nocase(input, "iso_8859_15") == 0
		   || strcmp_nocase(input, "iso-8859_15") == 0
		   || strcmp_nocase(input, "iso-8859-15") == 0) {
		charset = kCSISO_8859_15;  /**< ISO-8859-15. */
		return true;
	} else if (strcmp_nocase(input, "iso_8859-16") == 0
		   || strcmp_nocase(input, "iso_8859_16") == 0
		   || strcmp_nocase(input, "iso-8859_16") == 0
		   || strcmp_nocase(input, "iso-8859-16") == 0) {
		charset = kCSISO_8859_16;  /**< ISO-8859-16. */
		return true;
	} else {
		return false;
	}
}

bool charset2string(eCharsets charset, std::string& output)
{
	switch (charset) {
	case kCSASCII:
		output = "ascii";
		return true;
		break;
	case kCSISO_8859_1:
		output = "iso-8859-1";
		return true;
		break;
	case kCSISO_8859_8:
		output = "iso-8859-8";
		return true;
		break;
	case kCSUTF8:
		output = "utf8";
		return true;
		break;
	case kCSISO_8859_2:
		output = "iso-8859-2";
		return true;
		break;
	case kCSISO_8859_3:
		output = "iso-8859-3";
		return true;
		break;
	case kCSISO_8859_4:
		output = "iso-8859-4";
		return true;
		break;
	case kCSISO_8859_5:
		output = "iso-8859-5";
		return true;
		break;
	case kCSISO_8859_6:
		output = "iso-8859-6";
		return true;
		break;
	case kCSISO_8859_7:
		output = "iso-8859-7";
		return true;
		break;
	case kCSISO_8859_9:
		output = "iso-8859-9";
		return true;
		break;
	case kCSISO_8859_10:
		output = "iso-8859-10";
		return true;
		break;
	case kCSISO_8859_13:
		output = "iso-8859-13";
		return true;
		break;
	case kCSISO_8859_14:
		output = "iso-8859-14";
		return true;
		break;
	case kCSISO_8859_15:
		output = "iso-8859-15";
		return true;
		break;
	case kCSISO_8859_16:
		output = "iso-8859-16";
		return true;
		break;
	default:
		output = "";
		return false;
	}
}





/** Convert a string to an eBackendKind enumeration.
 *
 * Find out which eBackendKind enumeration constant most closely matches
 * a given string.
 *
 * @param input The string-representation (probably from the user)
 * which is to be converted to eBackendKind.
 *
 * @param backend The output backend kind.  Is only defined if the
 * function returns true.
 *
 * @return True iff we were able to recognize and thus convert the
 * backend kind.
 *
 */
bool string2backend_kind(const std::string& input, eBackendKind& backend)
{
	if (strcmp_nocase(input, "pg") == 0
	    || strcmp_nocase(input, "p") == 0
	    || strcmp_nocase(input, "postgresql") == 0
	    || strcmp_nocase(input, "postgres") == 0) {
		backend = kPostgreSQL;
		return true;
	} else if (strcmp_nocase(input, "my") == 0
		   || strcmp_nocase(input, "m") == 0
		   || strcmp_nocase(input, "mysql") == 0) {
		backend = kMySQL;
		return true;
	} else if (strcmp_nocase(input, "l") == 0
		   || strcmp_nocase(input, "2") == 0
		   || strcmp_nocase(input, "lt") == 0
		   || strcmp_nocase(input, "sqlite") == 0
		   || strcmp_nocase(input, "sqlite2") == 0
		   || strcmp_nocase(input, "s") == 0) {
		backend = kSQLite2;
		return true;
	} else if (strcmp_nocase(input, "3") == 0
		   || strcmp_nocase(input, "lt3") == 0
		   || strcmp_nocase(input, "sqlite3") == 0
		   || strcmp_nocase(input, "s3") == 0) {
		backend = kSQLite3;
		return true;
	} else {
		return false;
	}
}


/** Split a string into a list of strings, based on a string of
 * characters to split on.
 *
 * The characters are not included in the split strings.
 *
 * If none of the characters are found, then the list containing
 * instring is returned in outlist.
 *
 * @param instring The input string.
 *
 * @param splitchars The string of characters on which to split.
 *
 * @param outlist The output list of strings.*/
void split_string(const std::string& instring, const std::string& splitchars, std::list<std::string>& outlist)
{
	if (instring.find_first_not_of(splitchars, 0) == std::string::npos) {
		// Add instring as sole output in list
		outlist.push_back(instring);
		return;
	} else {
		std::string::size_type nbegin, nend;
		nbegin = instring.find_first_not_of(splitchars, 0);
		nend = instring.find_first_of(splitchars, nbegin);
		if (nend == std::string::npos) {
			outlist.push_back(instring.substr(nbegin, instring.length()-nbegin));
		} else {
			outlist.push_back(instring.substr(nbegin, nend-nbegin));
			nbegin = instring.find_first_not_of(splitchars, nend);
			while (nbegin != std::string::npos
			       && nend != std::string::npos) {
				nend = instring.find_first_of(splitchars, nbegin);
				if (nend == std::string::npos) {
					outlist.push_back(instring.substr(nbegin, instring.length()-nbegin));
					break;
				} else {
					outlist.push_back(instring.substr(nbegin, nend-nbegin));
				}
				nbegin = instring.find_first_not_of(splitchars, nend);
			}
		}
	}
}


/** Split a string into a list of strings, based on a substring to split on.
 *
 * The substring is not included in the split strings.
 *
 * If the substring is not found, then the list containing only
 * instring is returned in outlist.
 *
 * @param instring The input string.
 *
 * @param splitstring The string on which to split.
 *
 * @param outlist The output list of strings.*/
void split_string_on_string(const std::string& instring, const std::string& splitstring, std::list<std::string>& outlist)
{
	if (instring.find(splitstring, 0) == std::string::npos) {
		// Add instring as sole output in list
		outlist.push_back(instring);
		return;
	} else {
		std::string::size_type nbegin, nend;
		if (instring.find(splitstring, 0) == 0) {
			nbegin = splitstring.length();
		} else {
			nbegin = 0;
		}
		nend = instring.find(splitstring, nbegin + splitstring.length());
		if (nend == std::string::npos) {
			outlist.push_back(instring.substr(nbegin, instring.length()-nbegin));
		} else {
			outlist.push_back(instring.substr(nbegin, nend-nbegin));
			nbegin = nend + splitstring.length();
			while (nbegin != std::string::npos
			       && nend != std::string::npos
			       && nbegin != instring.length()) {
				nend = instring.find(splitstring, nbegin);
				if (nend == std::string::npos) {
					outlist.push_back(instring.substr(nbegin, instring.length()-nbegin));
					break;
				} else {
					outlist.push_back(instring.substr(nbegin, nend-nbegin));
				}
				nbegin = nend + splitstring.length();
			}
		}
	}
}


/** Split a string into a vector of string, based on a string of
 *  characters to split on.
 *
 * The characters are not included in the split strings.
 *
 * If none of the characters are found, then the list containing only
 * instring is returned in outlist.
 *
 * This just calls the list equivalent, and converts the list to a
 * vector.
 *
 * @param instring The input string.
 *
 * @param splitchars The string of characters on which to split.
 *
 * @param outlist The output vector of strings.*/
void split_string(const std::string& instring, const std::string& splitchars, std::vector<std::string>& outlist)
{
	std::list<std::string> result_list;
	split_string(instring, splitchars, result_list);
	outlist.resize(result_list.size());
	std::list<std::string>::const_iterator ci = result_list.begin();
	std::list<std::string>::const_iterator cend = result_list.end();
	std::vector<std::string>::size_type i = 0;
	while (ci != cend) {
		outlist[i] = *ci;
		++i;
		++ci;
	}
}

/** Split a string into a vector of string, based on a string 
 * to split on.
 *
 * The split-string is not included in the split strings.
 *
 * If the split-string is not found, then the list containing only
 * instring is returned in outlist.
 *
 * This just calls the list equivalent, and converts the list to a
 * vector.
 *
 * @param instring The input string.
 *
 * @param splitstring The string of characters on which to split.
 *
 * @param outlist The output vector of strings.*/
void split_string_on_string(const std::string& instring, const std::string& splitstring, std::vector<std::string>& outlist)
{
	std::list<std::string> result_list;
	split_string_on_string(instring, splitstring, result_list);
	outlist.resize(result_list.size());
	std::list<std::string>::const_iterator ci = result_list.begin();
	std::list<std::string>::const_iterator cend = result_list.end();
	std::vector<std::string>::size_type i = 0;
	while (ci != cend) {
		outlist[i] = *ci;
		++i;
		++ci;
	}
}

/** Remove from the beginning of a substring and to the end of the string.
 *
 * Removes everything in a string starting from the beginning of a
 * given substring to the end of the string.  If it is not found, then
 * the input string is returned.
 *
 *@param instr The input string (is not changed).
 *
 *@param substr The substring to find.
 *
 *@return The resulting string.
 */
std::string remove_from_substring_and_out(const std::string& instr, const std::string& substr)
{
	std::string::size_type pos = instr.find(substr);
	if (pos == std::string::npos) {
		return instr;
	} else if (pos == 0) {
		return "";
	} else {
		return instr.substr(0,pos);
	}
}


/** Convert a string containing C escapes to the string with the 
 * escapes translated to chraracters.
 *
 * @param instring The string to convert.
 *
 * @return instring, with C escape sequences converted to their 
 * character equivalents.
 */
std::string parse_C_backslashes(const std::string& instring)
{
	std::string result;
	std::string::size_type size = instring.length();
	for (std::string::size_type i = 0;
	     i < size;
	     ++i) {
		char c = instring[i];
		if (c == '\\') {
			if (i < size-1) {
				++i;
				char c2 = instring[i];
				switch (c2) {
				case 'n':
					result += '\n';
					break;
				case 't':
					result += '\t';
					break;
				case 'v':
					result += '\v';
					break;
				case 'b':
					result += '\b';
					break;
				case 'a':
					result += '\a';
					break;
				case 'r':
					result += '\r';
					break;
				case 'f':
					result += '\f';
					break;
				case '\\':
					result += '\\';
					break;
				case '?':
					result += '\?';
					break;
				case '\'':
					result += '\'';
					break;
				case '"':
					result += '"';
					break;
				case 'x':
					if (i < size-3) {
						std::string hex = std::string(1, instring[i+1]) + instring[i+2];
						if (is_hex(hex)) {
							i += 2;
							result += hex2char(hex.c_str());
						} else {
							result += c2;
						}
					} else {
						result += c2;
					}
					break;
				case '0':
				case '1':
				case '2':
				case '3':
					if (i < size-3) {
						if (isdigit(instring[i])
						    && isdigit(instring[i+1])
						    && isdigit(instring[i+2])) {
							std::string oct = std::string(1, instring[i]) + std::string(1, instring[i+1]) + instring[i+2];
							i += 2;
							result += octal2char(oct.c_str());
						} else {
							result += c2;
						}
					} else {
						result += c2;
					}
					break;
				default:
					result += c2;
					break;
				}
			} else {
				result += c;
			}
		} else {
			result += c;
		}
	}
	return result;
}



/** Replace a substring with another string in a string.
 *
 * @param instring the input string
 *
 * @param substring_to_replace the substring of instring to replace
 *
 * @param replacement_string the string to replce substring_to_replace by
 *
 * @return instring, with all non-overlapping instances of
 * substring_to_replace replaced by replacement_string.
 */
std::string replace_substring(const std::string& instring, std::string substring_to_replace, std::string replacement_string)
{
	if (instring.find(substring_to_replace, 0) == std::string::npos
	    || substring_to_replace.empty()) {
		return instring; // Return string without changing it
	} else if (instring == substring_to_replace) {
		return replacement_string;
	} else {
		std::string result;
		std::string::size_type nbegin, nend;
		if ((nbegin = instring.find(substring_to_replace, 0)) == 0) {
			// Start next search just after substring_to_replace
			nbegin = substring_to_replace.length();
			// Replace [0:len(substring_to_replace)] 
			// with replacement_string
			result = replacement_string;
		} else {
			// Add original chars up to the beginning
			// of the substring to replace
			result = instring.substr(0, nbegin);
		}
		nend = instring.find(substring_to_replace, nbegin);
		if (nend == std::string::npos) {
			result += instring.substr(nbegin);
		} else {
			result += instring.substr(nbegin, nend-nbegin);
			result += replacement_string;
			nbegin = nend + substring_to_replace.length();
			while (nbegin != std::string::npos
			       && nend != std::string::npos
			       && nbegin != instring.length()) {
				nend = instring.find(substring_to_replace, nbegin);
				if (nend == std::string::npos) {
					result += instring.substr(nbegin);
					break;
				} else {
					result += instring.substr(nbegin, nend-nbegin);
					result += replacement_string;
				}
				nbegin = nend + substring_to_replace.length();
			}
		}
		return result;
	}
}

/** Strip characters from begining and end of a string.
 *
 * @param instring The input string.
 *
 * @param stripchars The string containing the characters to
 * strip. Must not be empty.
 *
 * @return The result string, that is, instring with any characters
 * from stripchars removed from the beginning and end.
 */
std::string strip(const std::string& instring, std::string stripchars)
{
	std::string result;
	std::string::size_type i = instring.find_first_not_of(stripchars);
	if (i == std::string::npos) {
		return "";
	}
	result = instring.substr(i, std::string::npos);
	i = result.find_last_not_of(stripchars);
	if (i == std::string::npos) {
		return result;
	} else {
		return result.substr(0, i+1);
	}
}



/** Strip whitespace from beginning and end of instring.
 *
 * Removes all whitespace characters from beginning and end of
 * instring.
 *
 * @param instring The string to strip.
 *
 * @return The string with whitespace removed from both ends.
 */
std::string strip(const std::string& instring)
{
	return strip(instring, " \n\r\v\t");
}

std::string long2compact_string7bits(long input)
{
	std::string result;
	result.reserve(5);
	char m0,m1,m2,m3,m4;
	m0 = input & 0x7F;
	m1 = (input >> 7) & 0x7f;
	m2 = (input >> 14) & 0x7f;
	m3 = (input >> 21) & 0x7f;
	m4 = (input >> 28) & 0x7f;
	bool bBegun = false;
	if ((m4 > 0) || bBegun) {
		result += (m4 + '0');
		bBegun = true;
	}
	if ((m3 > 0) || bBegun) {
		result += (m3 + '0');
		bBegun = true;
	}
	if ((m2 > 0) || bBegun) {
		result += (m2 + '0');
		bBegun = true;
	}
	if ((m1 > 0) || bBegun) {
		result += (m1 + '0');
	}
	result += (m0 + '0');
	return result;
}


void compact_string7bits2long(const std::string& inStr, long& output)
{
	std::string::size_type i = 0, end = inStr.size();
	output = 0;
	if (end == 0) {
		return;
	}
	char c = inStr[i++];
	output |= c - '0';
	while (i != end) {
		c = inStr[i++];
		output <<= 7;
		output |= c - '0';
	}
}

void IntegerList2compact_string7bits(const IntegerList& intlist, std::string& result)
{
	result = " ";
	IntegerListConstIterator cit = intlist.const_iterator();
	while (cit.hasNext()) {
		long curlong = cit.next();
		std::string this_compact_string;
		result += long2compact_string7bits(curlong);
	}
	result += " ";
}

void compact_string7bits2IntegerList(const std::string& inStr, IntegerList& intlist)
{
	std::string::size_type i = 1; // We must skip the initial space
	std::string::size_type end = inStr.size();
	if (end == 0) {
		return;
	}
	--end; // We must skip the last space.
	char c;
	while (i != end) {
		long cur_long = 0;
		bool bContinue = true;
		while (bContinue) {
			c = inStr[i++];
			if (c == ' ') {
				bContinue = false;
			} else {
				cur_long <<= 7;
				cur_long |= c - '0';
			}
			if (i == end) {
				bContinue = false;
			}
		}
		intlist.addIntegerBack(cur_long);
	}
}

/** Check whether input string ends with end string.
 *
 * @param input The string to test
 *
 * @param end The string which may or may not be at the end of the
 * input string.
 *
 * @return True if input ends with end , false otherwise.
 */
bool string_ends_with(const std::string& input, const std::string& end)
{
	// Does not work with UTF-8 in end...
	std::string::size_type input_length = input.length();
	std::string::size_type end_length = end.length();
	if (end_length > input_length) {
		return false;
	} else {
		std::string end_string = input.substr(input_length - end_length , end_length);
		return end_string == end;
	}
}


/** Convert a single Unicode character (a long) to UTF-8.
 *
 * "Stolen" from utf.c in SQLite 3, which is Public Domain code.
 *
 * @param input the Unicode character
 *
 * @param output The input encoded as a single UTF-8 character.
 */
void long2utf8(long c, std::string& output)
{
	output.clear();
	if (c<0x00080) {
		output += (c&0xFF);
	} else if (c<0x00800) {
		output += (0xC0 + ((c>>6)&0x1F));
		output += (0x80 + (c & 0x3F));
	} else if (c<0x10000) {
		output += (0xE0 + ((c>>12)&0x0F));
		output += (0x80 + ((c>>6) & 0x3F));
		output += (0x80 + (c & 0x3F));
	} else {
		output += (0xF0 + ((c>>18) & 0x07));
		output += (0x80 + ((c>>12) & 0x3F));
		output += (0x80 + ((c>>6) & 0x3F));
		output += (0x80 + (c & 0x3F));
	}
}

// Convert all non-identifier characters to underscore..
std::string convert_to_identifier(const std::string& instr)
{
	std::string result;
	
	std::string::size_type instr_size = instr.size();
	for (std::string::size_type i = 0; i < instr_size; ++i) {
		char c = instr[i];
		if ((c >= 'A' && c <= 'Z')
		    || (c >= 'a' && c <= 'z')
		    || (c >= '0' && c <= '9')
		    || (c == '_')) {
			result += c;
		} else {
			result += '_';
		}
	}

	// Prepend an underscore if it starts with a number.
	if (!result.empty()) {
		char first_char = result[0];
		if (first_char >= '0'
		    && first_char <= '9') {
			result = std::string("_") + result;
		}
	}

	for (int index = 0; emdros_reserved_words[index] != NULL; ++index) {
		if (strcmp_nocase(result, emdros_reserved_words[index]) == 0) {
			result = result + std::string("_");
			break;
		}
	}

	return result;
}

