/*
 * mql_select_statements.cpp
 *
 * MQL SELECT statements
 *
 * Ulrik Petersen
 * Created: 3/6-2001 (March 6, 2001)
 * Last update: 3/29-2007
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2007  Ulrik Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

/**@file mql_select_statements.cpp
 *@brief Implementation of MQL SELECT statements (MQL layer)
 */


#include <emdfdb.h>
#include <mql_result.h>
#include <mql_select_statements.h>
#include <mql_query.h>
#include <string_func.h>
#include <mql_R.h>
#include <logging.h>
#include <monads.h>
#include <inst.h>


//////////////////////////////////////////////////////////////
//
// QueryBase
//
//////////////////////////////////////////////////////////////
QueryBase::QueryBase(eFocusSpec focus, 
		     MQLMonadSetElement* in_spec, 
		     sheaf_return_type_pair *sheaf_return_type,
		     Topograph* query) 
	: m_focus(focus),
	  m_in(in_spec),
	  m_sheaf_return_type(sheaf_return_type),
	  m_query(query),
	  m_pArbitrary_monad_set(0)
{
}

QueryBase::QueryBase(eFocusSpec focus, 
		     std::string *pArbitrary_monad_set,
		     sheaf_return_type_pair *sheaf_return_type,
		     Topograph* query) 
	: m_focus(focus),
	  m_in(0),
	  m_sheaf_return_type(sheaf_return_type),
	  m_query(query),
	  m_pArbitrary_monad_set(pArbitrary_monad_set)
{
}

QueryBase::~QueryBase()
{
	delete m_in;
	delete m_query;
	delete m_sheaf_return_type->second;
	delete m_sheaf_return_type;
	delete m_pArbitrary_monad_set;
}

void QueryBase::myWeed(MQLExecEnv *pEE, bool& bResult)
{
	m_query->weed(pEE, bResult);
}

bool QueryBase::mySymbol(MQLExecEnv *pEE, bool& bResult)
{
	// Assume everything went well
	bResult = true;

	// Check arbitrary monad set existence, if necessary
	if (m_pArbitrary_monad_set != 0) {
		bool bExists;
		if (!pEE->pDB->monadSetExists(*m_pArbitrary_monad_set,
					      bExists,
					      m_Su)) {
			// DB error
			pEE->pError->appendError("Database error getting arbitrary monad set '" + *m_pArbitrary_monad_set + "'.");
			return false;
		} else {
			if (!bExists) {
				bResult = false;
				pEE->pError->appendError("Arbitrary monad set '" + *m_pArbitrary_monad_set + "' does not exist.");
				return true;
			}
		}
	}

	// If this is a flat sheaf, check the list of object type names
	if (m_sheaf_return_type->first == kFlatSheaf) {
		StringListConstIterator ci = m_sheaf_return_type->second->const_iterator();
		while (ci.hasNext()) {
			const std::string& object_type_name = ci.next();
      
			bool bDBResult;
			bool bObjectTypeExists;
			id_d_t object_type_id;
			eObjectRangeType dummyObjectRangeType;
			eMonadUniquenessType dummyMonadUniquenessType;
			bDBResult = pEE->pDB->objectTypeExists(object_type_name,
							       bObjectTypeExists, 
							       object_type_id,
							       dummyObjectRangeType,
							       dummyMonadUniquenessType);


			// Was there a DB error?
			if (!bDBResult) {
				pEE->pError->appendError("DB error checking existence of object type name '" + object_type_name + "' in flat sheaf list.\n");
				return false;
			}

			// Was the object type there?
			if (!bObjectTypeExists) {
				pEE->pError->appendError("Object type name '" + object_type_name + "' in flat sheaf list does not exist.\n");
				return false;
			}

			// If we came this far, it existed, 
			// so add it to the list
			m_flat_sheaf_object_types.push_back(std::pair<id_d_t, std::string>(object_type_id, object_type_name));
		}
	}

	return m_query->symbol(pEE, bResult);
}

bool QueryBase::myType(MQLExecEnv *pEE, bool& bResult)
{
	// FIXME: We should give kORTSingleRange if m_Su is a single range.
	return m_query->type(pEE, kORTMultipleRange, bResult);
}

bool QueryBase::myMonads(MQLExecEnv *pEE, bool& bResult)
{
	// Build m_Su
	if (m_in != 0)
		m_in->monadsBuildSet(m_Su);

	//
	// restrict to min_m .. max_m
	//

	// Get all_m-1
	bool bAllM1IsEmptyDatabase = false;
	try {
		if (!pEE->pDB->getAll_m_1(pEE->m_all_m_1)) {
			return false;
		}
	} catch (BadMonadsException e) {
		bAllM1IsEmptyDatabase = true;
	}

	// restrict Su
	if (bAllM1IsEmptyDatabase) {
		m_Su = SetOfMonads();
	} else {
		m_Su = SetOfMonads::intersect(m_Su, pEE->m_all_m_1);
	}

	// If we got this far, there were no DB errors
	return true;
}

bool QueryBase::myExec(MQLExecEnv *pEE, MQLResult **ppMQLResult)
{
	// If substrate is empty, we cannot retrieve anything.
	// We will even throw BadMonadsException if we go on.
	if (m_Su.isEmpty()) {
		pEE->pOBBVec = 0;

		// Declare sheaf
		Sheaf* pSheaf;

		// Write to log
		LOG_WRITE_TIME("QueryBase::myExec()", "Not executing because substrate is empty");

		// Create failed sheaf
		pSheaf = new Sheaf(true);

		// Make new MQLResult from sheaf
		*ppMQLResult = new MQLResult(pSheaf, pEE);
	} else {
		pEE->pOBBVec = new OBBVec;

		// Calculate multi-maps
		m_query->calculateMMap(pEE->pDB);

		// Perform aggregate querying
		if (!m_query->aggregateQuery(pEE, m_Su))
			return false;

		// Add copies of each object block's ObjectBlockBase to pEE->pOBBVec.
		// These will be used instead of the object block
		// when constructing the matched object
		m_query->addOBBToVec(pEE->pOBBVec);

		// Declare variables
		Sheaf* pSheaf;

		// Make universe
		SetOfMonads U(m_Su.first(), m_Su.last());

		// Write to log
		LOG_WRITE_TIME("QueryBase::myExec()", "R_topograph starting.");

		// Execute query
		try {
			pSheaf = R_topograph(pEE, U, m_Su, m_query);
		} catch (EMdFDBException e) {
			return false;
		}

		// Write to log
		LOG_WRITE_TIME("QueryBase::myExec()", "R_topograph finished.");

		// Make new MQLResult from sheaf
		*ppMQLResult = new MQLResult(pSheaf, pEE);
	}

	// If this is a flat sheaf, then flatten the sheaf
	if (m_sheaf_return_type->first == kFlatSheaf) {
		(*ppMQLResult)->flatten(m_flat_sheaf_object_types);
	}

	// If we got this far, there were no DB errors
	return true;
}


/////////////////////////////////////////////////////////////////
//
// SELECT (focus|all|) OBJECTS
//
// SelectStatement
//
/////////////////////////////////////////////////////////////////
SelectStatement::SelectStatement(MQLExecEnv *pEE, 
				 eFocusSpec focus, 
				 MQLMonadSetElement* in_spec, 
				 sheaf_return_type_pair *sheaf_return_type,
				 Topograph* query) 
	: Statement(pEE), QueryBase(focus, in_spec, sheaf_return_type, query)
{
}

SelectStatement::SelectStatement(MQLExecEnv *pEE, 
				 eFocusSpec focus, 
				 std::string* arbitrary_monad_set,
				 sheaf_return_type_pair *sheaf_return_type,
				 Topograph* query) 
	: Statement(pEE), QueryBase(focus, arbitrary_monad_set, sheaf_return_type, query)
{
}

SelectStatement::~SelectStatement()
{
}

void SelectStatement::weed(bool& bResult)
{
	// Just call base-class
	QueryBase::myWeed(m_pEE, bResult);
}

bool SelectStatement::symbol(bool& bResult)
{
	// Just call base-class
	return QueryBase::mySymbol(m_pEE, bResult);
}

bool SelectStatement::type(bool& bResult)
{
	// Just call base-class
	return QueryBase::myType(m_pEE, bResult);
}

bool SelectStatement::monads(bool& bResult)
{
	// Just call base-class
	return QueryBase::myMonads(m_pEE, bResult);
}

bool SelectStatement::exec()
{
	// Just call base-class
	return QueryBase::myExec(m_pEE, &m_result);
}

/////////////////////////////////////////////////////////////////
//
// SELECT OBJECTS AT
//
// SelectObjectsAtStatement
//
/////////////////////////////////////////////////////////////////
SelectObjectsAtStatement::SelectObjectsAtStatement(MQLExecEnv *pEE, monad_m at_monad, std::string* object_type_name) : Statement(pEE)
{
	m_at_monad = at_monad;
	m_object_type_name = object_type_name;
	m_object_type_id = NIL;
}

SelectObjectsAtStatement::~SelectObjectsAtStatement()
{
	delete m_object_type_name;
}

//
// - Check that the object type exists
//
bool SelectObjectsAtStatement::symbol(bool& bResult)
{
	// - Check that the object type exists
	eObjectRangeType dummyObjectRangeType;
	eMonadUniquenessType dummyMonadUniquenessType;
	if (!m_pEE->pDB->objectTypeExists(*m_object_type_name,
					  bResult,
					  m_object_type_id,
					  dummyObjectRangeType,
					  dummyMonadUniquenessType))
		return false;
	if (!bResult) {
		m_pEE->pError->appendError("Object type name '" + *m_object_type_name + "' does not exist.\n");
		return true;
	}
  
	// If we got this far, there were no DB errors and no compiler errors
	bResult = true;
	return true;
}

//
// - Just ask the database
//
bool SelectObjectsAtStatement::exec()
{
	// Get results from DB
	IDDList DBResult;
	if (!m_pEE->pDB->getObjectsStartingAtSm(*m_object_type_name, 
						m_at_monad, 
						DBResult))
		return false;

	// Create new MQLResult
	m_result = new MQLResult();
	m_result->appendHeader("id_d", kTCID_D);
  
	// Add the results from the DB to MQLResult
	IDDList::const_iterator cbegin(DBResult.begin());
	IDDList::const_iterator ci(cbegin);
	IDDList::const_iterator cend(DBResult.end());
	while (ci != cend) {
		// Add newline unless it is the first
		if (ci != cbegin)
			m_result->newline();

		// Append this id_d
		m_result->append(id_d2string(*ci));

		// Advance iterator
		++ci;
	}

	// If we got this far, there were no DB errors
	return true;
}


/////////////////////////////////////////////////////////////////
//
// SELECT OBJECTS HAVING MONADS IN
//
// SelectObjectsHavingMonadsInStatement
//
/////////////////////////////////////////////////////////////////
SelectObjectsHavingMonadsInStatement::SelectObjectsHavingMonadsInStatement(MQLExecEnv *pEE, 
									   MQLMonadSetElement* monads, 
									   const std::string& object_type_name) : ObjectTypesStatement(pEE, object_type_name), m_AST_monads(monads)
{
}

SelectObjectsHavingMonadsInStatement::~SelectObjectsHavingMonadsInStatement()
{
	delete m_AST_monads;
}

// 
// - Check that all the ranges of monads are positive and monotonic
//
void SelectObjectsHavingMonadsInStatement::weed(bool& bResult)
{
	// Monad-ranges should be positive and monotonic
	m_AST_monads->weed(m_pEE, true, bResult);
	if (!bResult)
		return;

	// If we got this far, there were no DB errors and no compiler errors
	bResult = true;
	return;
}

//
// - Build the set of monads
//
bool SelectObjectsHavingMonadsInStatement::monads(bool& bResult)
{
	//
	// - Build the set of monads
	//
	m_AST_monads->monadsBuildSet(m_monads);
  
	// There were no DB errors and no compiler errors
	bResult = true;
	return true;
} 


bool SelectObjectsHavingMonadsInStatement::exec()
{
	// Make result
	m_result = new MQLResult();
	m_result->appendHeader("object_type_name", kTCString);
	m_result->appendHeader("monad_m", kTCMonad_m);
	m_result->appendHeader("id_d", kTCID_D);

	// Call base class
	return ObjectTypesStatement::exec();
}

bool SelectObjectsHavingMonadsInStatement::execOneType(const ObjectTypeInfo& oti)
{
	// Get object type name
	std::string object_type_name = oti.getObjectTypeName();
	eObjectRangeType objectRangeType = oti.getObjectRangeType();

	// Ask DB
	if (!m_pEE->pDB->getObjectsHavingMonadsInMonadSet(object_type_name, 
							  m_monads,
							  objectRangeType,
							  *m_result)) {
		m_result->clear();
		return false;
	}

	// If we got this far, there were no DB errors
	return true;
}


  



/////////////////////////////////////////////////////////////////
//
// GET OBJECTS HAVING MONADS IN
//
// GetObjectsHavingMonadsInStatement
//
/////////////////////////////////////////////////////////////////
GetObjectsHavingMonadsInStatement::GetObjectsHavingMonadsInStatement(MQLExecEnv *pEE, 
								     MQLMonadSetElement* monads, 
								     std::string *object_type_name) 
	: ObjectTypeStatement(pEE, object_type_name), 
	  m_AST_monads(monads),
	  m_bGetAll(false),
	  m_pFeaturesToGet(0)
{
}

GetObjectsHavingMonadsInStatement::GetObjectsHavingMonadsInStatement(MQLExecEnv *pEE, 
								     MQLMonadSetElement* monads, 
								     std::string *object_type_name,
								     bool bDummy) 
	: ObjectTypeStatement(pEE, object_type_name), 
	  m_AST_monads(monads),
	  m_bGetAll(true),
	  m_pFeaturesToGet(0)
{
}

GetObjectsHavingMonadsInStatement::GetObjectsHavingMonadsInStatement(MQLExecEnv *pEE, 
								     MQLMonadSetElement* monads, 
								     std::string *object_type_name,
								     GrammarFeature *pFeaturesToGet) 
	: ObjectTypeStatement(pEE, object_type_name), 
	  m_AST_monads(monads),
	  m_bGetAll(false)
{
	if (pFeaturesToGet != 0) {
		m_pFeaturesToGet = Feature::FromGrammarFeature(pFeaturesToGet);
		delete pFeaturesToGet;
	}
}


GetObjectsHavingMonadsInStatement::~GetObjectsHavingMonadsInStatement()
{
	delete m_AST_monads;
	delete m_pFeaturesToGet;
}

// 
// - Check that all the ranges of monads are positive and monotonic
//
void GetObjectsHavingMonadsInStatement::weed(bool& bResult)
{
	// Monad-ranges should be positive and monotonic
	m_AST_monads->weed(m_pEE, true, bResult);
	if (!bResult)
		return;

	// If we got this far, there were no compiler errors
	bResult = true;
	return;
}


//
// - Check that the features exist for the given object type.
//
bool GetObjectsHavingMonadsInStatement::symbol(bool& bResult)
{
	if (!ObjectTypeStatement::symbolObjectTypeExists(bResult)) {
		return false;
	}
	if (!bResult) {
		return true;
	}

	// Is this ALL?
	if (m_bGetAll) {
		std::list<FeatureInfo> features_of_OT;
		if (!m_pEE->pDB->getFeaturesForObjectType(m_object_type_id,
							  features_of_OT)) {
			m_pEE->pError->appendError("Database error getting features for object type '" + *m_object_type_name + "'.\n");
			return false;
		}
		std::list<FeatureInfo>::reverse_iterator feat_rci = features_of_OT.rbegin();
		while (feat_rci != features_of_OT.rend()) {
			std::string featureName = feat_rci->getName();

			// We even add "self"
			Feature *pNewFeature = new Feature(new std::string(featureName), m_pFeaturesToGet);
			m_pFeaturesToGet = pNewFeature;

			feat_rci++;
		}
	}


	//
	if (m_pFeaturesToGet != 0) {
		m_pFeaturesToGet->assignListIndex(0);

		if (!m_pFeaturesToGet->symbolFeaturesExist(m_pEE, 
							   m_object_type_id, 
							   bResult))
			return false;
		if (!bResult)
			return true;
	}

	if (m_pFeaturesToGet != 0) {
		m_pObjectBlockBase = new ObjectBlockBase(new std::string(*m_object_type_name),
							 new std::string(""),
							 kRetrieve,
							 new Feature(*m_pFeaturesToGet),
							 false // Is (not, in this case) NOT EXIST
							 );
	} else {
		m_pObjectBlockBase = new ObjectBlockBase(new std::string(*m_object_type_name),
							 new std::string(""),
							 kRetrieve,
							 (Feature*) 0,
							 false // Is (not, in this case) NOT EXIST
							 );
	}
	m_pEE->pOBBVec = new OBBVec;
	m_pEE->pOBBVec->push_back(m_pObjectBlockBase);
  
	if (!m_pObjectBlockBase->symbol(m_pEE, bResult)) {
		return false;
	}
	if (!bResult) {
		return true;
	}


	// If we got this far, there were no DB errors and no compiler errors
	bResult = true;
	return true;
}

bool GetObjectsHavingMonadsInStatement::type(bool& bResult)
{
	if (!ObjectTypeStatement::type(bResult)) {
		return false;
	}
	if (!bResult) {
		return true;
	}

	//
	// - Get feature name and type
	//
	if (m_pFeaturesToGet != 0) {
		if (!m_pFeaturesToGet->typeFeatureName(m_pEE, bResult))
			return false;
		if (!bResult)
			return true;
	}

	// If we got this far, there were no DB errors and no compiler errors
	bResult = true;
	return true;
}


//
// - Build the set of monads
//
bool GetObjectsHavingMonadsInStatement::monads(bool& bResult)
{
	//
	// - Build the set of monads
	//
	m_AST_monads->monadsBuildSet(m_monads);
  
	// There were no DB errors and no compiler errors
	bResult = true;
	return true;
} 


bool GetObjectsHavingMonadsInStatement::exec()
{
	std::list<std::pair<id_d_t, std::string> > object_type_list;
	object_type_list.push_back(std::pair<id_d_t, std::string>(m_object_type_id, *m_object_type_name));

	Inst *pInst;
	std::vector<std::string> FeatureNames;
	std::list<FeatureInfo> FeatureInfos;
	if (m_pFeaturesToGet != 0) {
		m_pFeaturesToGet->execMakeFeatureNameVector(FeatureNames);
		m_pFeaturesToGet->execMakeFeatureList(FeatureInfos);
	}
	unsigned int no_of_features_to_get = FeatureNames.size();
	pInst = new Inst(FeatureNames);

	// Get all_m-1
	bool bAllM1IsEmptyDatabase = false;
	try {
		if (!m_pEE->pDB->getAll_m_1(m_pEE->m_all_m_1)) {
			return false;
		}
	} catch (BadMonadsException e) {
		bAllM1IsEmptyDatabase = true;
	}
  
	// Return an empty inst
	if (bAllM1IsEmptyDatabase) {
		delete pInst;
		ListOfStraws *pLOS = new ListOfStraws();
		Sheaf *pSheaf = new Sheaf(pLOS);

		m_result = new MQLResult(pSheaf, m_pEE);
		m_result->flatten(object_type_list);
		return true;
	}

	if (!m_pEE->pDB->getObjectsHavingMonadsIn(*m_object_type_name,
						  m_object_type_id,
						  m_objectRangeType,
						  m_monads,
						  m_pEE->m_all_m_1,
						  FeatureInfos,
						  pInst)) {
		delete pInst;
		Sheaf *pSheaf = new Sheaf(true); // A failed sheaf
		m_result = new MQLResult(pSheaf, m_pEE);
		m_result->flatten(object_type_list);
		return false;
	}


	Straw *pStraw = new Straw();
	SetOfMonads Universe = SetOfMonads(1, MAX_MONAD);
	pInst->set_current_universe(Universe);
	Inst::const_iterator ci = pInst->begin();
	while (ci != pInst->end()) {
		const InstObject* pInstObject = ci.next();

		id_d_t id_d = pInstObject->getID_D();

		SetOfMonads som;
		pInstObject->getMonads(som);

		MatchedObject *pMo = new MatchedObject(id_d,
						       som,
						       0, // inner sheaf
						       m_pObjectBlockBase);

		if (m_pFeaturesToGet != 0) {
			unsigned int feature_index = 0;
			for (feature_index = 0;
			     feature_index < no_of_features_to_get;
			     feature_index++) {
				const EMdFValue *pInstObjectValue = pInstObject->getFeature(feature_index);
				EMdFValue *pNewValue = new EMdFValue(*pInstObjectValue);
				pMo->addEMdFValue(feature_index, pNewValue);
			}
		}

		pStraw->append(pMo);
	}

	delete pInst;

	ListOfStraws *pLOS = new ListOfStraws();
	pLOS->append(pStraw);
	Sheaf *pSheaf = new Sheaf(pLOS);

	m_result = new MQLResult(pSheaf, m_pEE);
	m_result->flatten(object_type_list);

	// If we came this far, there were no DB errors
	return true;
}

  



/////////////////////////////////////////////////////////////////
//
// SELECT OBJECT TYPES
//
// SelectObjectTypesStatement
//
/////////////////////////////////////////////////////////////////
SelectObjectTypesStatement::SelectObjectTypesStatement(MQLExecEnv *pEE) : Statement(pEE)
{
}

SelectObjectTypesStatement::~SelectObjectTypesStatement()
{
}

bool SelectObjectTypesStatement::exec()
{
	// Ask database
	std::list<std::string> l;
	if (!m_pEE->pDB->getObjectTypes(l)) {
		return false;
	}

	// Create new MQLResult
	m_result = new MQLResult();
	m_result->appendHeader("object type", kTCString);
  
	// Add the results from the DB to MQLResult
	std::list<std::string>::const_iterator ci(l.begin());
	std::list<std::string>::const_iterator cend(l.end());
	while (ci != cend) {
		m_result->append(*ci);
		++ci;
		if (ci != cend)
			m_result->newline();
	}

	// Return no database error
	return true;
}



////////////////////////////////////////////////////////////////
//
// SELECT FEATURES
//
// SelectFeaturesStatement
//
///////////////////////////////////////////////////////////////
//
// - Check that the object type exists
//
bool SelectFeaturesStatement::symbol(bool& bResult)
{
	//
	// - Check that the object type exists
	//
	if (!symbolObjectTypeExists(bResult)) {
		if (!bResult) {
			m_pEE->pError->appendError("Object type name '" + *m_object_type_name + "' does not exist.\n");
			return true;
		}
	}

	//
	// If we got this far, there were no database errors
	return true;
}

bool SelectFeaturesStatement::exec()
{
	// Get result
	std::list<FeatureInfo> FeatureInfos;
	if (!m_pEE->pDB->getFeaturesForObjectType(m_object_type_id,
						  FeatureInfos))
		return false;

	// Create new MQLResult
	m_result = new MQLResult();
	m_result->appendHeader("feature name", kTCString);
	m_result->appendHeader("type", kTCString);
	m_result->appendHeader("default", kTCString);
	m_result->appendHeader("computed", kTCBool);

	// Convert result
	std::list<FeatureInfo>::const_iterator ci(FeatureInfos.begin());
	std::list<FeatureInfo>::const_iterator cend(FeatureInfos.end());
	while (ci != cend) {
		// Name
		m_result->append(ci->getName());

		// Type
		std::string strType_id;
		bool bTypeIDGottenOK = m_pEE->pDB->typeIdToString(ci->getType(), strType_id);
		m_result->append(strType_id);

		// Default value
		std::string defval;
		if (m_pEE->pOut->isXML()) {
			defval = ci->getDefaultValue();
		} else {
			defval = std::string("'") + ci->getDefaultValue() + "'";
		}
		m_result->append(defval);

		// Is_computed
		m_result->append(bool_alpha2string(ci->getIsComputed()));
		if (!bTypeIDGottenOK)
			return false;
		++ci;
		if (ci != cend)
			m_result->newline();
	}
  
	// If we got this far, there were no database errors
	return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT ENUMERATIONS
//
// SelectEnumerationsStatement
//
///////////////////////////////////////////////////////////////
SelectEnumerationsStatement::SelectEnumerationsStatement(MQLExecEnv *pEE) : Statement(pEE)
{
}

SelectEnumerationsStatement::~SelectEnumerationsStatement()
{
}

bool SelectEnumerationsStatement::exec()
{
	// Ask database
	std::list<std::string> l;
	if (!m_pEE->pDB->getEnumerations(l)) {
		return false;
	}

	// Create new MQLResult
	m_result = new MQLResult();
	m_result->appendHeader("enum name", kTCString);
  
	// Add the results from the DB to MQLResult
	std::list<std::string>::const_iterator ci(l.begin());
	std::list<std::string>::const_iterator cend(l.end());
	while (ci != cend) {
		m_result->append(*ci);
		++ci;
		if (ci != cend)
			m_result->newline();
	}

	// Return no database error
	return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT ENUMERATION CONSTANTS
//
// SelectEnumerationConstantsStatement
//
///////////////////////////////////////////////////////////////
bool SelectEnumerationConstantsStatement::symbol(bool& bResult)
{
	// Check that the enumeration exists and get its enum_id
	if (!symbolEnumerationExists(bResult, true)) 
		return false;

	// Return no database error
	return true;
}

bool SelectEnumerationConstantsStatement::exec()
{
	// Get result
	std::list<EnumConstInfo> l;
	if (!m_pEE->pDB->getEnumConstants(m_enum_id, l))
		return false;

	// Create new MQLResult
	m_result = new MQLResult();
	m_result->appendHeader("enum const", kTCString);
	m_result->appendHeader("value", kTCInteger);
	m_result->appendHeader("default", kTCBool);

	// Convert result
	std::list<EnumConstInfo>::const_iterator ci(l.begin());
	std::list<EnumConstInfo>::const_iterator cend(l.end());
	while (ci != cend) {
		m_result->append(ci->getName());
		m_result->append(int2string(ci->getValue()));
		m_result->append(bool_alpha2string(ci->getIsDefault()));
		++ci;
		if (ci != cend)
			m_result->newline();
	}
  
	// If we got this far, there were no database errors
	return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT OBJECT TYPES USING ENUMERATION
//
// SelectObjectTypesUsingEnumerationStatement
//
///////////////////////////////////////////////////////////////
bool SelectObjectTypesUsingEnumerationStatement::symbol(bool& bResult)
{
	// Check that the enumeration exists and get its enum_id
	if (!symbolEnumerationExists(bResult, true))
		return false;

	// Return no database error
	return true;
}

bool SelectObjectTypesUsingEnumerationStatement::exec()
{
	// Ask database
	std::list<std::string> l;
	if (!m_pEE->pDB->getObjectTypesUsingEnumeration(m_enum_id, l)) {
		return false;
	}

	// Create new MQLResult
	m_result = new MQLResult();
	m_result->appendHeader("object type name", kTCString);
  
	// Add the results from the DB to MQLResult
	std::list<std::string>::const_iterator ci(l.begin());
	std::list<std::string>::const_iterator cend(l.end());
	while (ci != cend) {
		m_result->append(*ci);
		++ci;
		if (ci != cend)
			m_result->newline();
	}

	// Return no database error
	return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT MAX_M
//
// SelectMaxMStatement
//
///////////////////////////////////////////////////////////////
bool SelectMaxMStatement::exec()
{
	// Ask database
	monad_m max_m;
	if (!m_pEE->pDB->getMax_m(max_m)) {
		return false;
	}

	// Create new MQLResult
	m_result = new MQLResult();
	m_result->appendHeader("max_m", kTCMonad_m);
  
	// Add the results from the DB to MQLResult
	m_result->startNewRow();
	m_result->append(monad_m2string(max_m));

	// Return no database error
	return true;
}

////////////////////////////////////////////////////////////////
//
// SELECT MIN_M
//
// SelectMinMStatement
//
///////////////////////////////////////////////////////////////
bool SelectMinMStatement::exec()
{
	// Ask database
	monad_m min_m;
	if (!m_pEE->pDB->getMin_m(min_m)) {
		return false;
	}

	// Create new MQLResult
	m_result = new MQLResult();
	m_result->appendHeader("min_m", kTCMonad_m);
  
	// Add the results from the DB to MQLResult
	m_result->startNewRow();
	m_result->append(monad_m2string(min_m));

	// Return no database error
	return true;
}


////////////////////////////////////////////////////////////////
//
// SELECT MONAD SETS
//
// SelectMonadSetsStatement
//
///////////////////////////////////////////////////////////////
bool SelectMonadSetsStatement::exec()
{
	std::list<std::string> monad_set_names;
	// Get all monad sets into m_monad_set_names
	if (!m_pEE->pDB->selectMonadSets(monad_set_names)) {
		m_pEE->pError->appendError("Could not get monad set names.\n");
		return false;
	}

	// Make result
	m_result = new MQLResult();
	m_result->appendHeader("monad_set_name", kTCString);

	std::list<std::string>::const_iterator ci;
	for (ci = monad_set_names.begin(); ci != monad_set_names.end(); ++ci) {
		m_result->startNewRow();
		m_result->append(*ci);
	}

	// If we got this far, there were no DB errors
	return true;
}


