/*
 * renderobjects.cpp
 *
 * Classes and functions to render Emdros objects into XML (or
 * something else), based on a JSON template.
 *
 * Ulrik Sandborg-Petersen
 * Created: 22/4-2007
 * Last update: 1/29-2009
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2007-2009  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

#include <renderobjects.h>
#include <string_func.h>
#include <mql_sheaf.h>
#include <emdf_value.h>
#include <algorithm>


#ifndef WIN32
#include <sys/time.h>
#include <time.h>
#endif



RenderObjects::RenderObjects(EmdrosEnv *pEnv, const std::string& db_name, const JSONValue *pFetchInfo, const std::string& stylesheet)
{
	m_pTemplateEE = new TemplateLangExecEnv();
	m_pTemplateEE->setDBName(db_name);

	m_db_name = db_name;
	m_pEE = pEnv;
	m_pFetchInfo = pFetchInfo;
	m_stylesheet = stylesheet;
	if (setDefaults()) {
		m_bIsOK = true;
		calculatePriorities();
	} else {
		m_bIsOK = false;
	}
}


RenderObjects::~RenderObjects()
{
	clean();
	delete m_pTemplateEE;

	String2PStringMap::iterator otnpsit = m_OTNPs.begin();
	while (otnpsit != m_OTNPs.end()) {
		delete otnpsit->second;
		++otnpsit;
	}
	m_OTNPs.clear();

	m_xml.clear();

	String2TemplateASTNodeMap::iterator templci;

	templci = m_startTemplates.begin();
	while (templci != m_startTemplates.end()) {
		delete (templci->second);
		++templci;
	}
	m_startTemplates.clear();

	templci = m_endTemplates.begin();
	while (templci != m_endTemplates.end()) {
		delete (templci->second);
		++templci;
	}
	m_endTemplates.clear();
}


bool RenderObjects::process(monad_m fm, monad_m lm)
{
	m_fm = fm;
	m_lm = lm;


	m_pTemplateEE->clean();

	if (!m_bIsOK) {
		return false;
	}

	std::cerr << "UP250: RenderObjects::process()" << std::endl;

#ifndef WIN32
	timeval t1, t3, t4, t5;
	gettimeofday(&t1, NULL);
#endif

	startDocument();

#ifndef WIN32
	gettimeofday(&t3, NULL);
	long int mytime = (t3.tv_usec + 1000000L * t3.tv_sec) - (t1.tv_usec + 1000000L * t1.tv_sec);
	std::cerr << "UP250: startDocument(): " << mytime << " microseconds." << std::endl;
#endif
	
	if (!readObjects()) {
		return false;
	}

#ifndef WIN32
	gettimeofday(&t4, NULL);
	mytime = (t4.tv_usec + 1000000L * t4.tv_sec) - (t3.tv_usec + 1000000L * t3.tv_sec);
	std::cerr << "UP250: readObjects(): " << mytime << " microseconds." << std::endl;
#endif


	objects2Document();

#ifndef WIN32
	gettimeofday(&t5, NULL);
	mytime = (t5.tv_usec + 1000000L * t5.tv_sec) - (t4.tv_usec + 1000000L * t4.tv_sec);
	std::cerr << "UP250: objects2Document(): " << mytime << " microseconds." << std::endl;
	mytime = (t5.tv_usec + 1000000L * t5.tv_sec) - (t1.tv_usec + 1000000L * t1.tv_sec);
	std::cerr << "UP250: whole RenderObjects process: " << mytime << " microseconds." << std::endl;
#endif
	return true;
}


std::string RenderObjects::getDocument()
{
	std::string newxml = joinListEmptyBetween(m_xml);
	if (m_pStyleSheet->hasObjectKey("postprocess")) {
		const JSONValue *pPostprocessValue = (*m_pStyleSheet)["postprocess"];
		if (pPostprocessValue->getKind() == kJSONList) {
			const std::list<const JSONValue*>& postprocess_list 
				= pPostprocessValue->getList();
			std::list<const JSONValue*>::const_iterator ci;
			ci = postprocess_list.begin();
			while (ci != postprocess_list.end()) {
				const JSONValue *pPostprocessObject
					= *ci;
				if (pPostprocessObject->getKind() == kJSONObject) {
					if (pPostprocessObject->hasObjectKey("instring")
					    && pPostprocessObject->hasObjectKey("outstring")) {
						std::string instring = (*pPostprocessObject)["instring"]->getString();
						std::string outstring = (*pPostprocessObject)["outstring"]->getString();
						newxml = replace_substring(newxml, instring, outstring);

		 			} else if (pPostprocessObject->hasObjectKey("prefix")) {
						std::string uncompiled_prefix = (*pPostprocessObject)["prefix"]->getString();

						bool bCompileResult = false;
						TemplateASTNode *pNode = parseTemplate(m_pTemplateEE, uncompiled_prefix, bCompileResult);
						if (!bCompileResult) {
							appendError("Error: parsing template 'prefix':\nparsing of template failed with the following error:\n" + m_pTemplateEE->getError());

						} else {
							pNode->exec(m_pTemplateEE);
							std::string real_prefix = m_pTemplateEE->getOutput();
							m_pTemplateEE->clearOutput();
							
							newxml = real_prefix + newxml;
						}

					} else if (pPostprocessObject->hasObjectKey("suffix")) {
						std::string uncompiled_suffix = (*pPostprocessObject)["suffix"]->getString();

						bool bCompileResult = false;
						TemplateASTNode *pNode = parseTemplate(m_pTemplateEE, uncompiled_suffix, bCompileResult);
						if (!bCompileResult) {
							appendError("Error: parsing template 'suffix':\nparsing of template failed with the following error:\n" + m_pTemplateEE->getError());

						} else {
							pNode->exec(m_pTemplateEE);
							std::string real_suffix = m_pTemplateEE->getOutput();
							m_pTemplateEE->clearOutput();
							
							newxml = newxml + real_suffix;
						}

					} else {
						std::cerr << "UP271: JSON object for 'postprocess' fetchinfo stylesheet member\ndoes not have either 'prefix' or 'suffix'." << std::endl;
					}
				} else {
					std::cerr << "UP270: JSON value for 'postprocess' fetchinfo stylesheet member\nis not a JSON Object or a JSON String." << std::endl;

				}
				++ci;
			}
		}
	}
	return "<?xml version='1.0' encoding='utf-8'?>" + newxml;
}



void RenderObjects::startDocument()
{
	m_xml.clear();
	// Done in getDOcument() instead
	// m_xml.push_back("<?xml version='1.0' encoding='utf-8' ?>");
}


void RenderObjects::calculatePriorities()
{
	int start_index = 0;
	int length = (int) m_OTHandlingPriorityList.size();
	ListOfString::const_iterator ci = m_OTHandlingPriorityList.begin();
	while (ci != m_OTHandlingPriorityList.end()) {
		std::string OTN = *ci;
		int end_index = length - start_index;

		m_startPriority.insert(std::make_pair(OTN, start_index));
		m_endPriority.insert(std::make_pair(OTN, end_index));

		++start_index;
		++ci;
	}
}


bool RenderObjects::readObjects()
{
	String2PStringMap::const_iterator ci = m_OTNPs.begin();
	SetOfMonads som(m_fm, m_lm);

	while (ci != m_OTNPs.end()) {
		std::string OTN = ci->first;

		const JSONValue *pOTNObject = (*m_pObjectTypes)[OTN];

		const JSONValue *pFeaturesToGet = 0;
		if (pOTNObject->hasObjectKey("get")) {
			pFeaturesToGet = (*pOTNObject)["get"];
		} else {
			// leave it as 0
		}

		if (!loadObjectsFromMonads(OTN, som, pFeaturesToGet)) {
			return false;
		}

		++ci;
	}

	return true;
}


void RenderObjects::objects2Document()
{
        monad_m monad;
	for (monad = m_fm; monad <= m_lm; ++monad) {
		const ListOfID_D& first_id_ds = getLOIDDFromFMS(monad);
		startElements(first_id_ds, monad);

		const ListOfID_D& last_id_ds = getLOIDDFromLMS(monad);
		endElements(last_id_ds, monad);
	}
}


typedef std::map<int, std::list<MemObject*> > Priority2ObjectMap;
typedef std::map<monad_m, Priority2ObjectMap > Monad2priority2objectMapMap;

void RenderObjects::startElements(const ListOfID_D& id_d_list, monad_m monad)
{
	if (!id_d_list.empty()) {
		Monad2priority2objectMapMap startDict;
		
		ListOfID_D::const_iterator id_dci = id_d_list.begin();
		while (id_dci != id_d_list.end()) {
			id_d_t id_d = *id_dci;
			
			MemObject *pObj = m_objs[id_d];
			const std::string& OTN = pObj->getObjectType();
			monad_m last = pObj->last();
			
			int priority = m_startPriority[OTN];
			
			startDict[last][priority].push_back(pObj);
			
			++id_dci;
		}

		Monad2priority2objectMapMap::reverse_iterator lmci = startDict.rbegin();
		while (lmci != startDict.rend()) {
			Priority2ObjectMap& priorityMap = lmci->second;
			Priority2ObjectMap::iterator priorityci = priorityMap.begin();
			while (priorityci != priorityMap.end()) {
				std::list<MemObject*>& memobjlist = priorityci->second;
				std::list<MemObject*>::iterator moit = memobjlist.begin();
				while (moit != memobjlist.end()) {
					MemObject *pObj = *moit;
					
					startElement(pObj, pObj->getObjectType());
					
					++moit;
				}
				
				++priorityci;
			}
			++lmci;
		}
	}
}


void RenderObjects::endElements(const ListOfID_D& id_d_list, monad_m monad)
{
	if (id_d_list.empty()) {
		return;
	}

	Monad2priority2objectMapMap endDict;

	ListOfID_D::const_iterator id_dci = id_d_list.begin();
	while (id_dci != id_d_list.end()) {
		id_d_t id_d = *id_dci;

		MemObject *pObj = m_objs[id_d];
		const std::string& OTN = pObj->getObjectType();
		monad_m first = pObj->first();
		
		int priority = m_endPriority[OTN];

		endDict[first][priority].push_back(pObj);

		++id_dci;
	}

	Monad2priority2objectMapMap::reverse_iterator fmci = endDict.rbegin();
	while (fmci != endDict.rend()) {
		Priority2ObjectMap& priorityMap = fmci->second;
		Priority2ObjectMap::iterator priorityci = priorityMap.begin();
		while (priorityci != priorityMap.end()) {
			std::list<MemObject*>& memobjlist = priorityci->second;
			std::list<MemObject*>::iterator moit = memobjlist.begin();
			while (moit != memobjlist.end()) {
				MemObject *pObj = *moit;

				endElement(pObj, pObj->getObjectType());
				
				++moit;
			}

			++priorityci;
		}
		++fmci;
	}
}


void RenderObjects::startElement(MemObject *pObj, const std::string& OTN)
{
	TemplateASTNode *pTemplate = m_startTemplates[OTN];
	if (pTemplate != 0) {
		m_pTemplateEE->setObject(pObj);
		pTemplate->exec(m_pTemplateEE);
		m_xml.push_back(m_pTemplateEE->getOutput());
		m_pTemplateEE->clearOutput();
	}
}


void RenderObjects::endElement(MemObject *pObj, const std::string& OTN)
{
	TemplateASTNode *pTemplate = m_endTemplates[OTN];
	if (pTemplate != 0) {
		m_pTemplateEE->setObject(pObj);
		pTemplate->exec(m_pTemplateEE);
		m_xml.push_back(m_pTemplateEE->getOutput());
		m_pTemplateEE->clearOutput();
	}
}


bool RenderObjects::loadObjectsFromMonads(const std::string& OTN, const SetOfMonads& som, const JSONValue *pFeaturesToGetValue)
{
	std::string *pOTN = m_OTNPs[OTN];

	std::string getClause;
	int nNoOfFeaturesToGet;
	if (pFeaturesToGetValue != 0) {
		const std::list<const JSONValue*>& features_to_get_list =
			pFeaturesToGetValue->getList();
		nNoOfFeaturesToGet = features_to_get_list.size();
	} else {
		nNoOfFeaturesToGet = 0;
	}

	if (nNoOfFeaturesToGet > 0) {
		const std::list<const JSONValue*>& features_to_get_list =
			pFeaturesToGetValue->getList();

		// Join list with ',' in between.
		std::ostringstream getClauseStream;
		getClauseStream << " GET ";
		std::list<const JSONValue*>::const_iterator ci = 
			features_to_get_list.begin();
		if (ci != features_to_get_list.end()) {
			getClauseStream << (*ci)->getString();
			++ci;
		}
		while (ci != features_to_get_list.end()) {
			getClauseStream << ',' << (*ci)->getString();
			++ci;
		}
		getClause = getClauseStream.str();
	} else {
		getClause = "";
	}

	std::string query = "GET OBJECTS HAVING MONADS IN";
	query += som.toString();
	query += "[" + OTN + getClause + "]GO";

#ifndef WIN32
	timeval t1, t2, t3, t4;
	gettimeofday(&t1, NULL);
#endif
	


	bool bCompilerResult;
	if (!m_pEE->executeString(query, bCompilerResult, false, false)) {
		appendError("RenderObjects::loadObjectsFromMonads: DB error executing query: " + query);
		return false;
	}
	if (!bCompilerResult) {
		appendError("RenderObjects::loadObjectsFromMonads: Query made a bad compiler result:" + query);
		return false;
	}
	if (!m_pEE->isFlatSheaf()) {
		appendError("RenderObjects::loadObjectsFromMonads: Query did not return a flat sheaf: " + query);
		return false;
	}

#ifndef WIN32
	gettimeofday(&t2, NULL);
	long long mytime = (t2.tv_usec + 1000000L * t2.tv_sec) - (t1.tv_usec + 1000000L * t1.tv_sec);
	std::cerr << "UP260: executeString() " << mytime << " microseconds. for query '" << query << "'" << std::endl;
#endif


	FlatSheaf *pFlatSheaf = m_pEE->takeOverFlatSheaf();
	FlatSheafConstIterator flatSheafCI = pFlatSheaf->const_iterator();
	while (flatSheafCI.hasNext()) {
		FlatStraw *pFlatStraw = flatSheafCI.next();
		FlatStrawConstIterator flatStrawCI = pFlatStraw->const_iterator();
		while (flatStrawCI.hasNext()) {
			MatchedObject *pMO = flatStrawCI.next();

			id_d_t id_d = pMO->getID_D();

			SetOfMonads mosom(pMO->getMonads());
			monad_m first = mosom.first();
			monad_m last = mosom.last();

			MemObject *pObj = new MemObject(nNoOfFeaturesToGet);
			pObj->setFirstLast(first, last);
			pObj->setObjectType(pOTN);
			pObj->setID_D(id_d);
			if (nNoOfFeaturesToGet > 0) {
				int index = 0;
				while (index < nNoOfFeaturesToGet) {
					pObj->setFeature(index, pMO->getFeatureAsString(index));

					++index;
				}
			}

			//m_objs.insert(std::make_pair(id_d, pObj));
			
			// The subscript[] operator is faster than
			// insert()!
			m_objs[id_d] = pObj;
			addToFMSandLMS(first, last, id_d);
		}
	}

#ifndef WIN32
	gettimeofday(&t3, NULL);
	mytime = (t3.tv_usec + 1000000L * t3.tv_sec) - (t2.tv_usec + 1000000L * t2.tv_sec);
	std::cerr << "UP260: getting objects from flat sheaf: " << mytime << " microseconds." << std::endl;
#endif
	

	delete pFlatSheaf;

#ifndef WIN32
	gettimeofday(&t4, NULL);
	mytime = (t4.tv_usec + 1000000L * t4.tv_sec) - (t3.tv_usec + 1000000L * t3.tv_sec);
	std::cerr << "UP260: deleting flat sheaf: " << mytime << " microseconds." << std::endl;
	mytime = (t4.tv_usec + 1000000L * t4.tv_sec) - (t1.tv_usec + 1000000L * t1.tv_sec);
	std::cerr << "UP260: whole loadObjectsFromSOM process: " << mytime << " microseconds." << std::endl;
#endif

	return true;
}


void RenderObjects::clean()
{
	ID_D2MemObjectMap::iterator it = m_objs.begin();
	while (it != m_objs.end()) {
		delete it->second;
		++it;
	}
	m_objs.clear();

	m_fms.clear();
	m_lms.clear();

	m_error_message = "";
}


bool RenderObjects::setDefaults()
{
	if (m_pFetchInfo->getKind() != kJSONObject) {
		appendError("fetchinfo is not a JSON object.");
		return false;
	}

	if (!m_pFetchInfo->hasObjectKey(m_stylesheet)) {
		appendError(std::string("fetchinfo does not have key '") + m_stylesheet + "'.");
		return false;
	}

	m_pStyleSheet = (*m_pFetchInfo)[m_stylesheet];

	bool bResult = true;

	if (!m_pStyleSheet->hasObjectKey("priority_list")) {
		appendError(std::string("stylesheet does not have key 'priority_list'"));
		return false;
	}

	const JSONValue *pPriorityList = (*m_pStyleSheet)["priority_list"];

	if (!m_pStyleSheet->hasObjectKey("object_types")) {
		appendError(std::string("stylesheet does not have key 'object_types'."));
		return false;
	}

	m_pObjectTypes = (*m_pStyleSheet)["object_types"];

	if (m_pObjectTypes->getKind() != kJSONObject) {
		appendError(std::string("stylesheet's \"object_types\" value is not a JSON Object, as it should be."));
		return false;
	}

	if (pPriorityList->getKind() != kJSONList) {
		appendError(std::string("stylesheet's \"prioritylist\" value is not a list, as it should be."));
		return false;
	}

	const std::list<const JSONValue*>& priority_list = pPriorityList->getList();
	std::list<const JSONValue*>::const_iterator plci = priority_list.begin();
	while (plci != priority_list.end()) {
		const JSONValue *pOTValue = *plci;

		if (pOTValue->getKind() != kJSONString) {
			std::ostringstream ostr;
			pOTValue->pretty(&ostr, 0); // The 0 is the indent level.
			appendError(std::string("stylesheet's \"prioritylist\" has a value which is not a JSON String.\nThe value looks like this:\n") + ostr.str());
			bResult = false;
		} else {
			std::string OTN = pOTValue->getString();

			m_OTHandlingPriorityList.push_back(OTN);

			m_OTNPs.insert(std::make_pair(OTN, new std::string(OTN)));

		
			// Handle m_pObjectTypes JSON Object
			if (!m_pObjectTypes->hasObjectKey(OTN)) {
				appendError("Error: \"object_types\" object does not have object type name\n\"" + OTN + "\", even though it is in \"priority_list\".");
				bResult = false;
			} else {
				const JSONValue *pOTNObject = (*m_pObjectTypes)[OTN];
				
				std::string start;
				if (pOTNObject->hasObjectKey("start")) {
					const JSONValue *pStart = (*pOTNObject)["start"];
					if (pStart->getKind() != kJSONString) {
						appendError("Error: for object type '" + OTN + "', \"start\" does not point to a string.");
						bResult = false;
					} else {
						start = (*pOTNObject)["start"]->getString();
					}
				}
				
				if (start.empty()) {
					m_startTemplates.insert(std::make_pair<std::string, TemplateASTNode*>(OTN, 0));
				} else {
					bool bCompileResult;
					TemplateASTNode *pNode = parseTemplate(m_pTemplateEE, start, bCompileResult);
					if (!bCompileResult) {
						appendError("Error: parsing template 'start' for object type '" + OTN + "':\nparsing of template failed with the following error:\n" + m_pTemplateEE->getError());
						bResult = false;
					}
					m_startTemplates.insert(std::make_pair(OTN, pNode));
				}
				
				std::string end;
				if (pOTNObject->hasObjectKey("end")) {
					const JSONValue *pEnd = (*pOTNObject)["end"];
					if (pEnd->getKind() != kJSONString) {
						appendError("Error: for object type '" + OTN + "', \"end\" does not point to a string.");
						bResult = false;
					} else {
						end = (*pOTNObject)["end"]->getString();
					}
				}

				if (end.empty()) {
					m_endTemplates.insert(std::make_pair<std::string, TemplateASTNode*>(OTN, 0));
				} else {
					bool bCompileResult;
					TemplateASTNode *pNode = parseTemplate(m_pTemplateEE, end, bCompileResult);
					if (!bCompileResult) {
						appendError("Error: parsing template 'end' for object type '" + OTN + "':\nparsing of template failed with the following error:\n" + m_pTemplateEE->getError());
						bResult = false;
					}
					m_endTemplates.insert(std::make_pair(OTN, pNode));

				}

				const JSONValue *pGetList = 0;
				if (pOTNObject->hasObjectKey("get")) {
					pGetList = (*pOTNObject)["get"];
				}

				if (pGetList != 0) {
					if (pGetList->getKind() != kJSONList) {
						appendError("Error: for object type '" + OTN + "', \"get\" is not a list.");
						pGetList = 0;
						bResult = false;
					}
					if (pGetList != 0) {
						const std::list<const JSONValue*>& getlist = pGetList->getList();
						std::list<const JSONValue*>::const_iterator ci = getlist.begin();
						while (ci != getlist.end()) {
							const JSONValue *pGetString = *ci;
							if (pGetString->getKind() != kJSONString) {
								appendError("Error: for object type '" + OTN + "', not all members of the \"get\" list are strings.");
								bResult = false;
							}
							++ci;
						}

					}
				}
			}
		}
		++plci;
	}

	return bResult;
}

void RenderObjects::appendError(const std::string& error)
{
    m_error_message += error;
    m_error_message += '\n';
}
