/*
 * renderxml.cpp
 *
 * Classes and functions to render XML into something else,
 * based on a JSON template.
 *
 * Ulrik Sandborg-Petersen 
 * Created: 4/22-2007
 * Last update: 6/12-2009
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2007-2009  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/


#include <renderxml.h>
#include <templatelang_classes.h>

#include <fstream>
#include <sstream>
#include <list>

#include <string_func.h>
#include <qdxml.h>

#ifndef WIN32
#include <sys/time.h>
#endif




/*
 * The UTF8-code was adapted from the code found in utf.c from SQLite 3.
 * The original code was in the Public Domain.
 *
 * Adapted by Ulrik Petersen.
 *
 */

/*
** This table maps from the first byte of a UTF-8 character to the number
** of trailing bytes expected. A value '4' indicates that the table key
** is not a legal first byte for a UTF-8 character.
*/
static const unsigned char extra_utf8_bytes[256]  = {
/* 0xxxxxxx */
0, 0, 0, 0, 0, 0, 0, 0,     0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,     0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,     0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,     0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,     0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,     0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,     0, 0, 0, 0, 0, 0, 0, 0,
0, 0, 0, 0, 0, 0, 0, 0,     0, 0, 0, 0, 0, 0, 0, 0,

/* 10wwwwww */
4, 4, 4, 4, 4, 4, 4, 4,     4, 4, 4, 4, 4, 4, 4, 4,
4, 4, 4, 4, 4, 4, 4, 4,     4, 4, 4, 4, 4, 4, 4, 4,
4, 4, 4, 4, 4, 4, 4, 4,     4, 4, 4, 4, 4, 4, 4, 4,
4, 4, 4, 4, 4, 4, 4, 4,     4, 4, 4, 4, 4, 4, 4, 4,

/* 110yyyyy */
1, 1, 1, 1, 1, 1, 1, 1,     1, 1, 1, 1, 1, 1, 1, 1,
1, 1, 1, 1, 1, 1, 1, 1,     1, 1, 1, 1, 1, 1, 1, 1,

/* 1110zzzz */
2, 2, 2, 2, 2, 2, 2, 2,     2, 2, 2, 2, 2, 2, 2, 2,

/* 11110yyy */
3, 3, 3, 3, 3, 3, 3, 3,     4, 4, 4, 4, 4, 4, 4, 4,
};

/*
** This table maps from the number of trailing bytes in a UTF-8 character
** to an integer constant that is effectively calculated for each character
** read by a naive implementation of a UTF-8 character reader. The code
** in the READ_UTF8 macro explains things best.
*/
static const int extra_utf8_bits[] =  {
  0,
  12416,          /* (0xC0 << 6) + (0x80) */
  925824,         /* (0xE0 << 12) + (0x80 << 6) + (0x80) */
  63447168        /* (0xF0 << 18) + (0x80 << 12) + (0x80 << 6) + 0x80 */
};

/*
** If a UTF-8 character contains N bytes extra bytes (N bytes follow
** the initial byte so that the total character length is N+1) then
** masking the character with utf8_mask[N] must produce a non-zero
** result.  Otherwise, we have an (illegal) overlong encoding.
*/
static const int utf_mask[] = {
  0x00000000,
  0xffffff80,
  0xfffff800,
  0xffff0000,
};

#define READ_UTF8(zIn, c) { \
  int extra;                                           \
  c = (unsigned int) *(zIn)++;			       \
  extra = extra_utf8_bytes[c];                         \
  switch( extra ){                                     \
    case 4: c = (int)0xFFFD; break;                    \
    case 3: c = (c<<6) + *(zIn)++;                     \
    case 2: c = (c<<6) + *(zIn)++;                     \
    case 1: c = (c<<6) + *(zIn)++;                     \
    c -= extra_utf8_bits[extra];                       \
    if( (utf_mask[extra]&c)==0                         \
        || (c&0xFFFFF800)==0xD800                      \
        || (c&0xFFFFFFFE)==0xFFFE ){  c = 0xFFFD; }    \
  }                                                    \
}

std::string utf8ToEntity(const std::string& inHTML)
{
	std::ostringstream ostr;
	const unsigned char *zIn = (const unsigned char*) inHTML.c_str();
	while (*zIn != '\0') {
		int c = 0;
		READ_UTF8(zIn, c);
		if (c < 0x80) {
			ostr << (char) c;
		} else {
			ostr << "&#" << c << ";";
		}
	}
	return ostr.str();
}

class ElementInfo {
private:
	std::string m_tag;
	TemplateASTNode *m_pStartTemplate;
	TemplateASTNode *m_pEndTemplate;
	bool m_bHasAttribute;
	bool m_bNix;
public:
	ElementInfo();
	~ElementInfo();
	bool fill(TemplateLangExecEnv *pEE, const std::string& tag, const JSONValue* pStart, const JSONValue *pEnd, bool bNix, std::string& /* out */ strErrorOutput);
	TemplateASTNode *getStartTemplate(void) { return m_pStartTemplate; };
	TemplateASTNode *getEndTemplate(void) { return m_pEndTemplate; };
	bool getBNix(void) const { return m_bNix; };
	const std::string& getTag(void) const { return m_tag; };
	bool hasAttribute(void) const { return m_bHasAttribute; };
};

ElementInfo::ElementInfo()
{
	m_bHasAttribute = false;
	m_tag = "";
	m_pStartTemplate = 0;
	m_pEndTemplate = 0;
}

ElementInfo::~ElementInfo()
{
	delete m_pStartTemplate;
	delete m_pEndTemplate;
}

bool ElementInfo::fill(TemplateLangExecEnv *pEE, const std::string& tag, const JSONValue* pStart, const JSONValue *pEnd, bool bNix, std::string& /* out */ strErrorOutput)
{
	m_bNix = bNix;
	m_bHasAttribute = false;
	m_tag = tag;

	if (pStart != 0) {
		if (pStart->getKind() != kJSONString) {
			strErrorOutput += "ElementInfo::fill(): for tag '" + tag + "', start is not a JSON String.";
			return false;
		}
		    
		bool bCompileResult;
		std::string start = pStart->getString();
		TemplateASTNode *pNode = parseTemplate(pEE, start, bCompileResult);
		if (!bCompileResult) {
			strErrorOutput += "ElementInfo::fill(): parsing template 'start' for tag '" + tag + "' failed with the following error:\n" + pEE->getError() + '\n';
			delete pNode;
			return false;
		} else {
			m_pStartTemplate = pNode;
		}

		if (start.find("{{ attrib", 0) != std::string::npos) {
			m_bHasAttribute = true;
		}
	}

	if (pEnd != 0) {
		if (pEnd->getKind() != kJSONString) {
			strErrorOutput += "ElementInfo::fill(): for tag '" + tag + "', end is not a JSON String.";
			return false;
		}
		    
		bool bCompileResult;
		TemplateASTNode *pNode = parseTemplate(pEE, pEnd->getString(), bCompileResult);
		if (!bCompileResult) {
			strErrorOutput += "ElementInfo::fill(): parsing template 'end' for tag '" + tag + "' failed with the following error:\n" + pEE->getError() + '\n';
			delete pNode;
			return false;
		} else {
			m_pEndTemplate = pNode;
		}
	}
	return true;
}

class XML2HTMLDocHandler : public QDDocHandler {
protected:
	std::string m_chars;
	typedef std::list<std::string> ListOfHTMLString;
	typedef std::list<std::string> ListOfXMLTag;
	ListOfXMLTag m_nixing_stack;
	TemplateLangExecEnv *m_pTemplateEE;
	typedef std::map<std::string, ElementInfo*> String2ElementInfoMap;
	String2ElementInfoMap m_elements;
        bool parseStyleSheet(const JSONValue *pStyleSheet, const std::string& render_stylesheet, std::string& strError);
 public:
	ListOfHTMLString m_html;
        XML2HTMLDocHandler(const JSONValue *pRenderInfo, const std::string& render_stylesheet, bool& /* out */ bCompilationOK, std::string& /* out */ strError);
        virtual ~XML2HTMLDocHandler() {
        };
        virtual void startElement(const std::string& tag, const AttributeMap& attrs) throw(QDException);
        virtual void endElement(const std::string& tag) throw(QDException);
        virtual void startDocument(void) throw(QDException);
        virtual void endDocument(void) throw(QDException);
        virtual void text(const std::string& str) throw(QDException);
	std::string getOutput(void);
protected:
	void addText(const std::string& str, bool bSpaceAfter = false, monad_m monad = 0);
	void emit(const std::string& str);
	void handleChars(monad_m monad = 0);
};

XML2HTMLDocHandler::XML2HTMLDocHandler(const JSONValue *pRenderInfo, const std::string& render_stylesheet, bool&  /* out */ bCompilationOK, std::string& /* out */ strError)
{
	m_pTemplateEE = new TemplateLangExecEnv();

	bCompilationOK = parseStyleSheet(pRenderInfo, render_stylesheet, strError);
}


bool XML2HTMLDocHandler::parseStyleSheet(const JSONValue *pRenderInfo, const std::string& render_stylesheet, std::string& /* out */ strError)
{
	std::cerr << "UP201: XML2HTMLDocHandler::parseStyleSheet: render_stylesheet = " << render_stylesheet << std::endl;
	bool bResult = true;
	if (pRenderInfo->getKind() != kJSONObject) {
		strError += "XML2HTMLDocHandler::parseStyleSheet: pRenderInfo is not a JSON Object.";
		return false;
	}
		     
	if (!pRenderInfo->hasObjectKey(render_stylesheet)) {
		strError += "XML2HTMLDocHandler::parseStyleSheet: pRenderInfo does not have '" + render_stylesheet + "' key.";
		return false;
	}

	const JSONValue *pStyleSheet = (*pRenderInfo)[render_stylesheet];

	if (pStyleSheet->getKind() != kJSONObject) {
		strError += "XML2HTMLDocHandler::parseStyleSheet: pStyleSheet is not a JSON Object.";
		return false;
	}
	if (!pStyleSheet->hasObjectKey("elements")) {
		strError += "XML2HTMLDocHandler::parseStyleSheet: pStyleSheet does not have 'elements' key.";
		return false;
	}
	const JSONValue *pElements = (*pStyleSheet)["elements"];
	
	if (pElements->getKind() != kJSONObject) {
		strError += "XML2HTMLDocHandler::parseStyleSheet: 'elements' value is not a JSON Object.";
		return false;
	}

	if (!pStyleSheet->hasObjectKey("extends")) {
		; // Nothing to do
	} else {
		const JSONValue *pExtends = pStyleSheet->getObjectValue("extends");
		if (pExtends->getKind() == kJSONString) {
			std::string extends = pExtends->getString();
			bResult = parseStyleSheet(pRenderInfo, extends, strError);
		} else if (pExtends->getKind() == kJSONNull) {
			; // That's OK, it can be 'null'.
		} else {
			strError += "XML2HTMLDocHandler::parseStyleSheet: 'renderinfo'['" + render_stylesheet + "']['extends'] does not map to a JSON string.";
			bResult = false;
		}
	}

	std::list<std::string> elements_keys = pElements->getObjectKeys();
	std::list<std::string>::const_iterator elemkeyci = 
		elements_keys.begin();
	while (elemkeyci != elements_keys.end()) {
		std::string elemkey = *elemkeyci;

		std::cerr << "UP206: Handling element key:" << elemkey << std::endl;

		const JSONValue *pValue = (*pElements)[elemkey];

		String2ElementInfoMap::iterator it = m_elements.find(elemkey);
		if (it != m_elements.end()) {
			delete it->second;
			m_elements.erase(it);
		}

		if (pValue->getKind() != kJSONObject) {
			strError += "XML2HTMLDocHandler::parseStyleSheet: for elements member '"
				+ elemkey + "', value is not a JSON Object.";
			bResult = false;
		} else {
			const JSONValue *pStartValue = 0;
			const JSONValue *pEndValue = 0;
			
			if (pValue->hasObjectKey("start")) {
				pStartValue = (*pValue)["start"];
			}
			
			if (pValue->hasObjectKey("end")) {
				pEndValue = (*pValue)["end"];
			}

			bool bNix = false;
			if (pValue->hasObjectKey("nix")) {
				bNix = (*pValue)["nix"]->getBoolean();
			}
			
			ElementInfo *pElementInfo = new ElementInfo();
			
			if (!pElementInfo->fill(m_pTemplateEE,
						elemkey,
						pStartValue,
						pEndValue,
						bNix,
						strError)) {
				
				bResult = false;
				delete pElementInfo;
			} else {
				m_elements.insert(std::make_pair(elemkey, pElementInfo));
			}
			
			++elemkeyci;
		}
	}

	return bResult;
}



void XML2HTMLDocHandler::emit(const std::string& str)
{
	if (m_nixing_stack.empty()) {
		m_html.push_back(str);
	}
}



void XML2HTMLDocHandler::startElement(const std::string& tag, const AttributeMap& attrs) throw(QDException)
{
	handleChars();

	String2ElementInfoMap::iterator it = m_elements.find(tag);
	if (it != m_elements.end()) {
		ElementInfo *pElementInfo = it->second;
		bool bNix = pElementInfo->getBNix();
		if (bNix) {
			m_nixing_stack.push_back(tag);
		} else {
			TemplateASTNode *pTemplate = pElementInfo->getStartTemplate();
			if (pTemplate != 0) {
				if (pElementInfo->hasAttribute()) {
					m_pTemplateEE->setAttributes(attrs);
				}
				pTemplate->exec(m_pTemplateEE);
				emit(m_pTemplateEE->getOutput());	
				m_pTemplateEE->clearOutput();
			}
		}
	} else {
		throw QDException("Error: Unknown start-tag '<" + tag + ">'");
	}
}


void XML2HTMLDocHandler::endElement(const std::string& tag) throw(QDException)
{
	handleChars();

	String2ElementInfoMap::iterator it = m_elements.find(tag);
	if (it != m_elements.end()) {
		ElementInfo *pElementInfo = it->second;
		bool bNix = pElementInfo->getBNix();
		if (bNix) {
			std::string nixing_top = *m_nixing_stack.rbegin();
			m_nixing_stack.pop_back();
			if (nixing_top != tag) {
				throw QDException("Nixing stack pop didn't match push");
			}
		} else {
			TemplateASTNode *pTemplate = pElementInfo->getEndTemplate();
			if (pTemplate != 0) {
				pTemplate->exec(m_pTemplateEE);
				emit(m_pTemplateEE->getOutput());	
				m_pTemplateEE->clearOutput();
			}
		}
	} else {
		throw QDException("Error: Unknown start-tag '</" + tag + ">'");
	}

}


void XML2HTMLDocHandler::startDocument(void) throw(QDException)
{
}


void XML2HTMLDocHandler::endDocument(void) throw(QDException)
{
}


void XML2HTMLDocHandler::text(const std::string& str) throw(QDException)
{
	m_chars += str;
}

void XML2HTMLDocHandler::addText(const std::string& str, bool bSpaceAfter, monad_m monad)
{
	std::string mystr;
	mystr = replace_substring(str, "&", "&amp;");
	mystr = replace_substring(mystr, "\"", "&quot;");
	mystr = replace_substring(mystr, "<", "&lt;");
	mystr = replace_substring(mystr, ">", "&gt;");
	
	std::string result;
	if (monad != 0) {
		result = "<a name=\"m" + monad_m2string(monad) + "\">";
	}
	result += mystr;
	if (bSpaceAfter) {
		result += ' ';
	}
	if (monad != 0) {
		result += "</a>";
	}
	emit(result);
}



void XML2HTMLDocHandler::handleChars(monad_m monad)
{
	if (m_chars.empty()) {
		return; // Nothing to do
	} else if (!m_nixing_stack.empty()) {
		// Clean up for next text() call.
		m_chars = "";
		return;
	} else if (!is_other_than_whitespace(m_chars)) {
		//		if (m_bInParagraph) {
			addText("", true); // Empty string, with a space after it!
			//}
			
		// Clean up for next text() call.
		m_chars = "";
		return;
	} else {
		if (is_whitespace(m_chars[0])) {
			addText("", true); // Empty string, with a space after it!
		}
		std::list<std::string> token_list;
		split_string(m_chars, " \n\t\r", token_list);
		std::list<std::string>::const_iterator ci = token_list.begin();
		std::list<std::string>::const_iterator cend = token_list.end();
		while (ci != cend) {
			std::string text = *ci;
			++ci;
			bool bAddSpaceAfter = true;
			if (ci == cend) {
				if (is_whitespace(m_chars[m_chars.length()-1])) {
					; // Leave it as true
				} else {
					bAddSpaceAfter = false;
				}
			}
			addText(text, bAddSpaceAfter);
		}

		// Clean up for next text() call.
		m_chars = "";
	}
}



std::string XML2HTMLDocHandler::getOutput(void)
{
	std::string result(joinList("", m_html));
	return result;
}


std::string RenderXML(const std::string& xml, const JSONValue *pTopStyleSheet, const std::string& render_stylesheet, bool& bProcessOK, std::string& strError)
{
	bProcessOK = true;

	std::cerr << "UP200: RenderXML" << std::endl;

	ASSERT_THROW(pTopStyleSheet->getKind() == kJSONObject,
		     "RenderXML: Top stylesheet is not a JSON Object");
	ASSERT_THROW(pTopStyleSheet->hasObjectKey("renderinfo"),
		     "RenderXML: Top stylesheet does not have 'renderinfo' member.");
	const JSONValue *pRenderInfo = (*pTopStyleSheet)["renderinfo"];

	ASSERT_THROW(pRenderInfo->getKind() == kJSONObject,
		     "RenderXML: renderinfo stylesheet is not a JSON OBject");
	ASSERT_THROW(pRenderInfo->hasObjectKey(render_stylesheet),
		     "RenderXML: renderinfo stylesheet does not have '" + render_stylesheet + "' member.");

	const JSONValue *pStyleSheet = (*pRenderInfo)[render_stylesheet];

#ifndef WIN32
	timeval t1, t2, t3;
	gettimeofday(&t1, NULL);
#endif
	std::string newxml = xml;
	if (pStyleSheet->hasObjectKey("preprocess")) {
		const JSONValue *pPreprocessValue = (*pStyleSheet)["preprocess"];
		if (pPreprocessValue->getKind() == kJSONList) {
			const std::list<const JSONValue*>& preprocess_list 
				= pPreprocessValue->getList();
			std::list<const JSONValue*>::const_iterator ci;
			ci = preprocess_list.begin();
			while (ci != preprocess_list.end()) {
				const JSONValue *pPreprocessObject
					= *ci;
				if (pPreprocessObject->getKind() == kJSONObject) {
					if (pPreprocessObject->hasObjectKey("instring")
					    && pPreprocessObject->hasObjectKey("outstring")) {
						std::string instring = (*pPreprocessObject)["instring"]->getString();
						std::string outstring = (*pPreprocessObject)["outstring"]->getString();
						newxml = replace_substring(newxml, instring, outstring);

		 			} else if (pPreprocessObject->hasObjectKey("prefix")) {
						std::string prefix = (*pPreprocessObject)["prefix"]->getString();
						newxml = replace_substring(newxml, "?>", "?>" + prefix);

					} else if (pPreprocessObject->hasObjectKey("suffix")) {
						std::string suffix = (*pPreprocessObject)["suffix"]->getString();
						newxml = newxml + suffix;

					} else {
						std::cerr << "UP271: JSON object for 'preprocess' renderinfo stylesheet member\ndoes not have both 'instring' and 'outstring'." << std::endl;
					}
				} else {
					std::cerr << "UP270: JSON value for 'preprocess' renderinfo stylesheet member\nis not a JSON Object or a JSON String." << std::endl;

				}
				++ci;
			}
		}
	}
	std::istringstream istr(newxml);

#ifndef WIN32
	gettimeofday(&t2, NULL);
	long long mytime = (t2.tv_usec + 1000000L * t2.tv_sec) - (t1.tv_usec + 1000000L * t1.tv_sec);
	std::cerr << "UP200: replacing replikregiB" << mytime << " microseconds." << std::endl;
#endif

        XML2HTMLDocHandler dh(pRenderInfo, render_stylesheet, bProcessOK, strError);
	if (!bProcessOK) {
		return "";
	}

	std::string output_string;
	try {
		QDParser parser;
		parser.parse(&dh, &istr);
		output_string = dh.getOutput();
	} catch (QDException e) {
		strError += "UP100:" + e.what() + "\n";
		bProcessOK = false;
	}

	
	//std::cerr << "UP217: output_string = " << output_string << "'" << std::endl;

#ifndef WIN32
	gettimeofday(&t3, NULL);
	mytime = (t3.tv_usec + 1000000L * t3.tv_sec) - (t2.tv_usec + 1000000L * t2.tv_sec);
	std::cerr << "UP200: parsing: " << mytime << " microseconds." << std::endl;
	mytime = (t3.tv_usec + 1000000L * t3.tv_sec) - (t1.tv_usec + 1000000L * t1.tv_sec);
	std::cerr << "UP200: whole XML2HTML process: " << mytime << " microseconds." << std::endl;
#endif

	return output_string;
}

