/*
 * pennimporter.cpp
 *
 * Functions and classes for importing Penn Treebank-style data
 * Created: 2/18-2006
 * Last update: 2/5-2008
 *
 */
/************************************************************************
 *
 *   Emdros - the database engine for analyzed or annotated text
 *   Copyright (C) 2001-2008  Ulrik Sandborg-Petersen
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation, license version 2.  
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *   02111-1307 USA
 *
 *
 *   Special exception
 *   =================
 * 
 *   In addition, as a special exception, Ulrik Petersen, the
 *   copyright holder of Emdros, gives permission to link Emdros, in
 *   whole or in part, with the libraries which are normally
 *   distributed with:
 *   
 *   a) Sun's Java platform,
 *   b) Python, 
 *   c) Jython,
 *   d) Ruby, and/or 
 *   e) Perl 
 *   f) PostgreSQL
 *   g) OpenSSL
 *
 *   (or with modified versions of these), and to distribute linked
 *   combinations including both Emdros, in whole or in part, and one
 *   or more of the libraries normally distributed with (a)-(g) above.
 *
 *   Please note: This gives you special rights concerning the
 *   libraries which normally accompany the above pieces of software.
 *   It gives you no special rights concerning software that you write
 *   yourself.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than the libraries
 *   normally distributed with (a)-(g) above.
 *
 *   If you modify this file, you may extend this exception to your
 *   version of the file, but you are not obligated to do so. If you
 *   do not wish to do so, delete this exception statement from your
 *   version.
 *
 *
 *   Other licensing forms
 *   =====================
 *
 *   If you wish to negotiate commercial licensing, please contact
 *   Ulrik Petersen at ulrikp[at]users.sourceforge.net.
 *
 *   Licensing can also be negotiated if your organization is an
 *   educational, non-profit, charity, missionary or similar
 *   organization.
 *
 *
 *   Website
 *   =======
 *
 *   Emdros has a website here:
 *
 *   http://emdros.org
 *
 *
 *
 **************************************************************************/

/**@file pennimporter.cpp
 *@brief Implementation of a Penn Treebank importer.
 */

#include "pennparser.h"
#include <penn_yylex.h>
#include <pennimporter.h>
#include "pennschema.h"
#include <fstream>

// Prototypes
extern void *PennParserAlloc(void *(*mallocProc)(size_t));
extern void PennParserFree(void *p, void (*freeProc)(void*));
extern void PennParser(
		       void *yyp,                       /* The parser */
		       int yymajor,                     /* The major token code number */
		       PennToken *yyminor,              /* The value for the token */
		       PennImporterEnv *pArg            /* optional argument */
		       );

PennToken *newPennToken()
{
	PennToken *pToken = new PennToken;
	pToken->pString = 0;
	return pToken;
}

extern void PennParserTrace(FILE *TraceFILE, char *zTracePrompt);

int penn_yyparse(PennImporterEnv *pEE)
{
	void *pParser;
	PennToken *pToken;
	int hTokenId;

	// PennParserTrace(stdout, "TRACE: ");
  
	pEE->bSyntaxError = false,
		pParser = PennParserAlloc( malloc );
	pToken = newPennToken();
	while((hTokenId = penn_yylex(pToken, pEE))
	      && !pEE->bSyntaxError) {
	  
		PennParser(pParser, hTokenId, pToken, pEE);
		pToken = newPennToken();
	}
	PennParser(pParser, 0, pToken, pEE);
	PennParserFree(pParser, free );
	deletePennToken(pToken);
	if (pEE->bSyntaxError) {
		return 1;
	} else {
		return 0;
	}
}

bool EmdrosImporterBase::readDocumentsFromFiles(const std::list<std::string>& filename_list)
{
	std::list<std::string>::const_iterator ci, cend;
	ci = filename_list.begin();
	cend = filename_list.end();

	bool bResult = true;
	while (ci != cend
	       && bResult) {
		bResult = readDocumentFromFile(*ci);
		++ci;
	}
      
	return bResult;
}

bool EmdrosImporterBase::readDocumentFromFile(const std::string& filename)
{
	bool bResult = false;
	std::ifstream fin;
	fin.open(filename.c_str());
	if (!fin) {
		std::cerr << "FAILURE: Could not open file " << filename << " for reading." << std::endl;
		std::cerr << "FAILURE: Action aborted." << std::endl;
		bResult = false;
	} else {
		bResult = readDocumentFromStream(&fin);
	}

	return bResult;
}



void PennTreeNode::setNodeType(const std::string& strIn)
{
	if (strIn == "-NONE-") {
		m_strNodeType = strIn;
	} else if (strIn == "") {
		m_strNodeType = strIn;
	} else if (strIn.find_first_not_of("-#") == std::string::npos) {
		m_strNodeType = strIn; // For things like '--' in TIGER corpus
	} else {
		std::list<std::string> outlist;
		split_string(strIn, "-#", outlist);
		std::list<std::string>::size_type outlist_size = outlist.size();
		if (outlist_size <= 1) {
			m_strNodeType = strIn;
		} else {
			std::string last_element = *outlist.rbegin();
			if (string_is_number(last_element)) {
				m_strCoref = last_element;
				if (outlist_size == 2) {
					m_strNodeType = *outlist.begin();
				} else {
					m_strNodeType = *outlist.begin();
					
					//
					// Copy the ones between the first and the last
					//
					std::list<std::string>::const_iterator ci, cendminus1;
					
					// Get the last element
					cendminus1 = outlist.end();
					--cendminus1;
					
					// Get one past the beginning
					ci = outlist.begin();
					++ci;
					
					std::list<std::string> innerlist;
					while (ci != cendminus1) {
						innerlist.push_back(*ci);
						++ci;
					}
					
					m_strFunction = joinList("-", innerlist);
				}
			} else {
				m_strNodeType = *outlist.begin();

				//
				// Copy the ones from one past the first to the last
				//
				std::list<std::string>::const_iterator ci, cend;
				
				// Get the last element
				cend = outlist.end();
				
				// Get one past the beginning
				ci = outlist.begin();
				++ci;
				
				std::list<std::string> innerlist;
				while (ci != cend) {
					innerlist.push_back(*ci);
					++ci;
				}
				
				m_strFunction = joinList("-", innerlist);
			}
		}
	}
}

void PennTreeNode::setTerminal(const std::string& strIn)
{
	if (strIn.find_first_of("*") != std::string::npos) {
		if (strIn.find_first_of("-") != std::string::npos) {
			std::string::size_type last_index_dash = strIn.find_last_of("-");
			std::string supposedCoref = strIn.substr(last_index_dash+1);
			if (string_is_number(supposedCoref)) {
				m_strCoref = supposedCoref;
				m_strTerminal = strIn.substr(0,last_index_dash);
			} else {
				m_strTerminal = strIn;
			}
		} else {
			m_strTerminal = strIn;
		}
	} else {
		m_strTerminal = strIn;
	}
}


PennImporterEnv::~PennImporterEnv() 
{
	delete pLexer;
}

void PennImporterEnv::clean() 
{
	delete pLexer;
	pLexer = 0;
	strErrorMessage = "";
	bSyntaxError = false;
	pCorpus = 0;
}




PennTreebankImporter::PennTreebankImporter(monad_m start_monad, id_d_t start_id_d)
	: EmdrosImporterBase(start_monad, start_id_d),
	  m_pCorpus(0)
{
	m_pEE = new PennImporterEnv();
}


PennTreebankImporter::~PennTreebankImporter()
{
	delete m_pEE;
	delete m_pCorpus;
}


bool PennTreebankImporter::readDocumentFromStream(std::istream *pIn)
{
	bool bResult = false;
	m_pEE->pLexer = new PennScanner(pIn);
	int nResult = penn_yyparse(m_pEE);
	if (nResult != 0) {
		std::cerr << "FAILURE: Syntax error parsing Penn Treebank document\n"
			  << "FAILURE: Error message was: " << m_pEE->strErrorMessage << std::endl;
		delete m_pEE->pCorpus;
		m_pEE->pCorpus = 0;
		m_pEE->clean();
		bResult = false;
	} else {
		m_pCorpus = m_pEE->pCorpus;
		m_pEE->pCorpus = 0;
		m_pEE->clean();
		bResult = true;
	}

	if (bResult) {
		// Reserve id_d and first monad for doc
		id_d_t doc_id_d = m_cur_id_d;
		++m_cur_id_d;
		monad_m doc_first_monad = m_cur_monad;

		// Assign monads and id_ds
		PennTreeNodeList *pCorpus = m_pCorpus;
		PennTreeNodeListConstIterator corpus_CI = pCorpus->const_iterator();
		while (corpus_CI.hasNext()) {
			// Ugly, I know, but I'm lazy and don't want to 
			// implement ParserListIterator.
			PennTreeNode *pTopTree = const_cast<PennTreeNode*>(corpus_CI.next());
			SetOfMonads sibling_monads;
			assignMonadsAndID_Ds(pTopTree, sibling_monads, doc_id_d);
		}

		// Resolve coreferences
		CorefMap mymap;
		corpus_CI = pCorpus->const_iterator();
		while (corpus_CI.hasNext()) {
			// Ugly, I know, but I'm lazy and don't want to 
			// implement ParserListIterator.
			PennTreeNode *pTopTree = const_cast<PennTreeNode*>(corpus_CI.next());
			resolveCorefs(pTopTree, mymap);
		}
		
		// Create objects
		corpus_CI = pCorpus->const_iterator();
		while (corpus_CI.hasNext()) {
			// Ugly, I know, but I'm lazy and don't want to 
			// implement ParserListIterator.
			PennTreeNode *pTopTree = const_cast<PennTreeNode*>(corpus_CI.next());
			createObjects(pTopTree);
		}

		// Create document
		createDoc(doc_first_monad, m_cur_monad-1, doc_id_d);

		// Clean up...
		clearNodes();
	}

	return bResult;
}


void PennTreebankImporter::putSchema(std::ostream *pOut)
{
	(*pOut) << penn_treebank_schema << std::endl;
}

void PennTreebankImporter::putCorpusAsMQL(std::ostream *pOut)
{
	PEMOList::const_iterator ci, cend, ctmp;
	
	// Documents
	(*pOut) << "BEGIN TRANSACTION GO\n";
	(*pOut) << "////////////////////////////////\n"
		<< "//\n"
		<< "// Documents\n"
		<< "//\n"
		<< "////////////////////////////////\n";
	std::list<FeatureInfo> document_feature_infos; // This is empty!
	(*pOut) << "CREATE OBJECTS WITH OBJECT TYPE [Document]\n";
	for (ci = m_docs.begin();
	     ci != m_docs.end();
	     ++ci) {
		(*ci)->putMQL(pOut, document_feature_infos, "");
	}
	(*pOut) << "GO\n\n";


	// Roots
	(*pOut) << "////////////////////////////////\n"
		<< "//\n"
		<< "// Roots\n"
		<< "//\n"
		<< "////////////////////////////////\n";
	std::list<FeatureInfo> root_feature_infos; // This is empty!
	root_feature_infos.push_back(FeatureInfo("parent", 
						 FEATURE_TYPE_ID_D, 
						 NIL_AS_STRING, 
						 false));
	(*pOut) << "CREATE OBJECTS WITH OBJECT TYPE [Root]\n";
	for (ci = m_root_trees.begin();
	     ci != m_root_trees.end();
	     ++ci) {
		(*ci)->putMQL(pOut, root_feature_infos, "");
	}
	(*pOut) << "GO\n\n";


	// Nonterminals
	(*pOut) << "////////////////////////////////\n"
		<< "//\n"
		<< "// Nonterminals\n"
		<< "//\n"
		<< "////////////////////////////////\n";
	std::list<FeatureInfo> nonterminal_feature_infos;
	nonterminal_feature_infos.push_back(FeatureInfo("parent", 
						 FEATURE_TYPE_ID_D, 
						 NIL_AS_STRING, 
						 false));
	nonterminal_feature_infos.push_back(FeatureInfo("coref", 
						 FEATURE_TYPE_LIST_OF_ID_D, 
						 "()", 
						 false));
	nonterminal_feature_infos.push_back(FeatureInfo("mytype", 
						 FEATURE_TYPE_STRING, 
						 "", 
						 false));
	nonterminal_feature_infos.push_back(FeatureInfo("function", 
						 FEATURE_TYPE_STRING, 
						 "", 
						 false));
	bool bTransactionInProgress = true;
	(*pOut) << "CREATE OBJECTS WITH OBJECT TYPE [Nonterminal]\n";

	long counter = 1;
	for (ci = m_nonterminals.begin();
	     ci != m_nonterminals.end();
	     ++ci) {
		(*ci)->putMQL(pOut, nonterminal_feature_infos, "");
		++counter;
		if ((counter % 50000) == 1) {
			(*pOut) << "GO\nCOMMIT TRANSACTION GO\n";
			ctmp = ci;
			++ctmp;
			if (ctmp != m_nonterminals.end()) {
				(*pOut) << "BEGIN TRANSACTION GO\n";
				(*pOut) << "CREATE OBJECTS WITH OBJECT TYPE [Nonterminal]\n";
				bTransactionInProgress = true;
			} else {
				bTransactionInProgress = false;
			}
			
		}
	}
	(*pOut) << "GO\n";
	/*
	if (bTransactionInProgress) {
		(*pOut) << "COMMIT TRANSACTION GO\n\n";
	}
	*/


	// Tokens
	(*pOut) << "////////////////////////////////\n"
		<< "//\n"
		<< "// Tokens\n"
		<< "//\n"
		<< "////////////////////////////////\n";
	std::list<FeatureInfo> token_feature_infos;
	token_feature_infos.push_back(FeatureInfo("parent", 
						  FEATURE_TYPE_ID_D, 
						  NIL_AS_STRING, 
						  false));
	token_feature_infos.push_back(FeatureInfo("coref", 
						  FEATURE_TYPE_LIST_OF_ID_D, 
						  "", 
						  false));
	token_feature_infos.push_back(FeatureInfo("surface", 
						  FEATURE_TYPE_STRING, 
						  "", 
						  false));
	token_feature_infos.push_back(FeatureInfo("mytype", 
						  FEATURE_TYPE_STRING, 
						  "", 
						  false));
	token_feature_infos.push_back(FeatureInfo("function", 
						  FEATURE_TYPE_STRING, 
						  "", 
						  false));
	bTransactionInProgress = true;
	//(*pOut) << "BEGIN TRANSACTION GO\n";
	(*pOut) << "CREATE OBJECTS WITH OBJECT TYPE [Token]\n";

	//counter = 1;
	for (ci = m_terminals.begin();
	     ci != m_terminals.end();
	     ++ci) {
		(*ci)->putMQL(pOut, token_feature_infos, "");
		++counter;
		if ((counter % 50000) == 1) {
			(*pOut) << "GO\nCOMMIT TRANSACTION GO\n";
			ctmp = ci;
			++ctmp;
			if (ctmp != m_terminals.end()) {
				(*pOut) << "BEGIN TRANSACTION GO\n";
				(*pOut) << "CREATE OBJECTS WITH OBJECT TYPE [Token]\n";
				bTransactionInProgress = true;
			} else {
				bTransactionInProgress = false;
			}
			
		}
	}
	(*pOut) << "GO\n";
	if (bTransactionInProgress) {
		(*pOut) << "COMMIT TRANSACTION GO\n\n";
	}
}

void doAddCorefToMap(PennTreeNode *pNode, CorefMap& mymap)
{
	if (pNode->getStrCoref() != "") {
		long coref = string2long(pNode->getStrCoref());
		if (mymap.find(coref) == mymap.end()) {
			mymap[coref] = std::list<id_d_t>();
		}
		id_d_t myid_d = pNode->getID_D();
		mymap[coref].push_back(myid_d);
	} 
}

void doAddCorefsToNode(PennTreeNode *pNode, CorefMap& mymap)
{
	if (pNode->getStrCoref() != "") {
		long coref = string2long(pNode->getStrCoref());
		id_d_t myid_d = pNode->getID_D();
		const std::list<id_d_t>& mylist = mymap[coref];
		std::list<id_d_t>::const_iterator ci = mylist.begin();
		std::list<id_d_t>::const_iterator cend = mylist.end();
		while (ci != cend) {
			if (*ci != myid_d) {
				pNode->getCorefs().push_back(*ci);
			}
			++ci;
		}
	}
}

void addCorefsToMap(PennTreeNode *pNode, CorefMap& mymap)
{
	if (pNode->getType() == kPennTerminal) {
		doAddCorefToMap(pNode, mymap);
	} else { // kPennRoot and kPennNonTerminal
		doAddCorefToMap(pNode, mymap);

		if (pNode->getChild() != 0) {
			addCorefsToMap(pNode->getChild(), mymap);
		}

		PennTreeNode *pSiblingNode = pNode->getNextSibling();
		if (pSiblingNode != 0) {
			addCorefsToMap(pSiblingNode, mymap);
		}
	}
}

void addCorefsToNodes(PennTreeNode *pNode, CorefMap& mymap)
{
	if (pNode->getType() == kPennTerminal) {
		doAddCorefsToNode(pNode, mymap);
	} else { // kPennRoot and kPennNonTerminal
		doAddCorefsToNode(pNode, mymap);

		if (pNode->getChild() != 0) {
			addCorefsToNodes(pNode->getChild(), mymap);
		}

		PennTreeNode *pSiblingNode = pNode->getNextSibling();
		if (pSiblingNode != 0) {
			addCorefsToNodes(pSiblingNode, mymap);
		}
	}

}

void PennTreebankImporter::resolveCorefs(PennTreeNode *pNode, CorefMap& mymap)
{
	addCorefsToMap(pNode, mymap);
	addCorefsToNodes(pNode, mymap);
}

void PennTreebankImporter::createObjects(PennTreeNode *pNode)
{
	EmdrosMemObject *pObj = 0;
	if (pNode->getType() == kPennRoot) {
		pObj = createRootTree(pNode);
	} else if (pNode->getType() == kPennNonTerminal) {
		PennTreeNode *pChild = pNode->getChild();
		if (pChild != 0 && pChild->getType() == kPennTerminal) {
			pObj = createTerminal(pChild, pNode);
		} else {
			pObj = createNonTerminal(pNode);
		}
	} else {
		; // Skip over terminals!
	}

	if (pNode->getType() == kPennTerminal) {
		; // Do nothing further
	} else {
		if (pNode->getChild() != 0) {
			createObjects(pNode->getChild());
		}

		PennTreeNode *pSiblingNode = pNode->getNextSibling();
		if (pSiblingNode != 0) {
			createObjects(pSiblingNode);
		}
	}
}

EmdrosMemObject *PennTreebankImporter::createTerminal(PennTreeNode *pNode, PennTreeNode *pParent)
{
	EmdrosMemObject *pObj = new EmdrosMemObject(pParent->getMonads(), pParent->getID_D());
	pObj->setFeature("parent", id_d2string(pParent->getParentID_D()));
	if (!pNode->getCorefs().empty()) {
		pObj->setFeature("coref", std::string("(") + joinList(",", pNode->getCorefs()) + ")");
	}
	pObj->setFeature("surface", pNode->getTerminal());
	pObj->setFeature("mytype", pParent->getNodeType());
	pObj->setFeature("function", pParent->getFunction());
	m_terminals.push_back(pObj);
	return pObj;
}


EmdrosMemObject *PennTreebankImporter::createNonTerminal(PennTreeNode *pNode)
{
	EmdrosMemObject *pObj = new EmdrosMemObject(pNode->getMonads(), pNode->getID_D());
	pObj->setFeature("parent", id_d2string(pNode->getParentID_D()));
	if (!pNode->getCorefs().empty()) {
		pObj->setFeature("coref", std::string("(") + joinList(",", pNode->getCorefs()) + ")");
	}
	pObj->setFeature("mytype", pNode->getNodeType());
	pObj->setFeature("function", pNode->getFunction());
	m_nonterminals.push_back(pObj);
	return pObj;
}


EmdrosMemObject *PennTreebankImporter::createRootTree(PennTreeNode *pNode)
{
	EmdrosMemObject *pObj = new EmdrosMemObject(pNode->getMonads(), pNode->getID_D());
	pObj->setFeature("parent", id_d2string(pNode->getParentID_D()));
	m_root_trees.push_back(pObj);
	return pObj;
}

void PennTreebankImporter::clearNodes(void)
{
	delete m_pCorpus;
	m_pCorpus = 0;
}

EmdrosMemObject *PennTreebankImporter::createDoc(monad_m first, monad_m last, id_d_t id_d)
{
	SetOfMonads monads(first, last);
	EmdrosMemObject *pObj = new EmdrosMemObject(monads, id_d);
	m_docs.push_back(pObj);
	return pObj;
}

SetOfMonads PennTreebankImporter::assignMonadsAndID_Ds(PennTreeNode *pNode, SetOfMonads& sibling_monads, id_d_t parent_id_d)
{
	if (pNode->getType() == kPennTerminal) {
		// Don't increase m_cur_id_d here, 
		// since we won't create an object from the terminal.
		// Instead, we will create one object from this
		// and its parent NonTerminal.
		//
		// We therefore set the id_d of this node
		// to that of the parent NonTerminal.
		pNode->setID_D(parent_id_d);
		pNode->setParentID_D(parent_id_d);
		pNode->addMonad(m_cur_monad);
		++m_cur_monad;
		return pNode->getMonads();
	} else { // kRoot and kNonTerminal
		// Set id_d
		pNode->setID_D(m_cur_id_d);
		++m_cur_id_d;

		pNode->setParentID_D(parent_id_d);

		SetOfMonads child_monads;
		if (pNode->getChild() != 0) {
			SetOfMonads mysibling_monads;
			child_monads = assignMonadsAndID_Ds(pNode->getChild(), mysibling_monads, pNode->getID_D());
			pNode->addMonads(child_monads);
			pNode->addMonads(mysibling_monads);
		}

		PennTreeNode *pSiblingNode = pNode->getNextSibling();
		if (pSiblingNode != 0) {
			assignMonadsAndID_Ds(pSiblingNode, sibling_monads, parent_id_d);
			sibling_monads.unionWith(pSiblingNode->getMonads());
		}

		return pNode->getMonads();
	}
}
